import groovy.sql.Sql
import org.apache.log4j.Logger

import com.sas.oprisk.server.services.ServiceLocator
import com.sas.oprisk.server.behavior.SecurityUtils
import com.sas.oprisk.server.workflow.WorkflowUtil
import com.sas.oprisk.server.BusinessObjectTypeInfo
import com.sas.oprisk.server.LinkInstance
import com.sas.oprisk.server.AttachmentContainer;
import com.sas.oprisk.server.util.BusinessObjectUtils;
import com.sas.oprisk.server.workflow.WorkflowUtil;
import com.sas.oprisk.framework.server.objects.TaggedPointerHandle;


def logger = Logger.getLogger("com.sas.oprisk.server.PurgeObjectsGroovy");
logger.setLevel(org.apache.log4j.Level.DEBUG);

// get schema prefix to use with sql schema.   
def schema = ServiceLocator.getContentService().getSchemaPrefix();

// find out whether it's from a web url standalone test (true) or a data unloader (false).   
def fromWeb = (request != null) ? true : false;

def ps = binding.hasVariable('psession') ? psession : ServiceLocator.getPersistenceSession();
ps.putCachedValue("purgeObjects", true);

def sql = new Sql(ps.getSqlConnection());

// initialize the error message as empty as it's used to communicate back to unloader for its returning status.   
errorMsg = "";


try {
   
   /*
    *************************************************************************************
    **
    **   Utilities
    ** 
    *************************************************************************************
    */
   
   // prints a line with 1 line break.
   def pl = { msg -> if (fromWeb) {println "${msg}<br/>"}; logger.debug(msg); }

      
   // prints a line and second param is the number of line breaks.
   def plv = {printPayLoad, extraLineBreaks = 0 ->
      pl printPayLoad;
      extraLineBreaks.times {
          pl "";
      }
   }

   
   // execute a sql statement with parameters.
   def execSqlParams = {sqlStmt, params = null, n = 0 ->
      try {
         if (params != null) {
            plv "${sqlStmt} with params ${params}", n;
            return sql.executeUpdate(sqlStmt.toString(), params);
         } else {
            plv "${sqlStmt}", n;
            return sql.executeUpdate(sqlStmt.toString());            
         }
      } catch(Exception e) {
         logger.error(e.getMessage())
         plv e.getMessage(), n;
         throw e;
      }
   }

      
   // read the value of the specified colNm from the fisrt row of the results
   def readValue = {sqlStmt, colNm ->
      try {
         plv "${sqlStmt}", 1;
         def firstRow = sql.firstRow(sqlStmt.toString());
         return firstRow != null ? firstRow.get(colNm) : null;
      } catch(Exception e) {
         logger.error(e.getMessage())
         plv e.getMessage();
         throw e;
      }
   }
   

   // read all the values.
   def readRks = {sqlStmt, params, colNm, n = 0 ->
      def resultRks = [];
      try {
         plv "${sqlStmt} with params ${params} ----- read the values for the column '${colNm}'", n;
         def results = sql.rows(sqlStmt.toString(), params);
         results.each {
            resultRks << it.get(colNm).longValue();
         }
         return resultRks;
      } catch(Exception e) {
         logger.error(e.getMessage())
         plv e.getMessage();
         throw e;
      }
   }

   
   // convert between "customObject#" and "CUST_OBJ_#".
   def getCustomObjectNumber = {typeName ->
      def number = 0;
      if (typeName?.trim()) {
         def prefixArray = ["customObject", "CUST_OBJ_"];
         prefixArray.each {prefix ->
            if (typeName.startsWith(prefix)) {
               try {
                  number = Integer.parseInt(typeName.substring(prefix.length()));
               } catch (Exception e) {
                  number = 0;
               }
            }
         }
      }
      return number;
   }

   
   // get a list of link instance Rks.
   def getLinkInstanceRks = {pbo, boTypeName, objRk ->
      def liRks = [];
      if (pbo != null) {
         def linkInstances = LinkInstance.object.getLinkInstances(pbo, ps);
         linkInstances.each {linkInstance ->
            liRks << linkInstance.getLinkInstanceRk();
         }
      } else {
         def linkInstanceSqlStmt = """
            select distinct lInstVar.link_instance_rk from ${schema}link_instance lInstVar, ${schema}LINK_TYPE_L lTypeVar 
            where (lInstVar.LINK_TYPE_RK = lTypeVar.LINK_TYPE_RK) and 
            (
               ((lTypeVar.BUSINESS_OBJECT_TYPE_NM_1 = ?) and (lInstVar.BUSINESS_OBJECT_RK_1 = ?)) 
                  or 
               ((lTypeVar.BUSINESS_OBJECT_TYPE_NM_2 = ?) and (lInstVar.BUSINESS_OBJECT_RK_2 = ?))
            )
            """;
         liRks = readRks(linkInstanceSqlStmt, [boTypeName, objRk, boTypeName, objRk], 'link_instance_rk', 1);
      }
      return liRks;
   }

      
   // get a list of attachment Rks.
   def getAttachmentRks = {pbo, busObjNm, objRk ->
      def attachmentRks = [];      
      if (pbo != null) {
         def attachments = pbo.getAttachments(ps);
         attachments.each {attachment ->
            attachmentRks << attachment.getAttachmentRk();
         }
      } else {
         def attSqlStmt = "select distinct attachment_rk from ${schema}attachment where business_object_nm = ? and business_object_rk = ?";
         attachmentRks = readRks(attSqlStmt, [busObjNm, objRk], 'attachment_rk', 1);
      }
      return attachmentRks;
   }

   
   // split a list into an equal sized lists
   def splitToBuckets = {array, size ->
      def partitions = [];
      int partitionCount = array.size() / size;
  
      partitionCount.times {partitionNumber ->
          def start = partitionNumber * size;
          def end = start + size - 1;
          partitions << array[start..end];
      }
  
      if (array.size() % size) {
         partitions << array[partitionCount * size..-1];
      }
       
      return partitions;
   }
   
   
   
   /*
    *************************************************************************************
    **
    **   Main
    **
    *************************************************************************************
    */

   
   //----------------------------------------------------------------------------------
   // check superuser first to make sure that only superuser can run this groovy code.      
   //----------------------------------------------------------------------------------
   def currentUser = binding.hasVariable('currentUser') ? currentUser : SecurityUtils.getUser(ps);   
   if (!currentUser.isSuperUser(ps)) {
      errorMsg = "Only administrator can run this tool for purging objects.";
      plv errorMsg
      return "fail"
   }
   
   def obj = null;
   
   // fetch boRk and boType in case this groovy code is running from the web url (while it's usually just a testing case).
   Long boRk = fromWeb ? Long.valueOf(request.getParameter("boRk")) : null;   
   def boType = fromWeb ? request.getParameter("boType") : null;

   Long objRk = binding.hasVariable('businessObjectRk') ? (Long)businessObjectRk : boRk;
   def busObjNm = binding.hasVariable('businessObjectNm') ? businessObjectNm : boType; // busObjNm is like "CUST_OBJ_#
   def boTypeName = "customObject" + getCustomObjectNumber(busObjNm); // boTypeName is like "customObject#"
   
   if (fromWeb) {
      //-----------------------------------
      // From Web request - test mode only
      //-----------------------------------      
      if (boRk == null || boType == null) {
         errorMsg = "Cannot retrieve businessObject from boType=${boType} / boRk=${boRk}";
         plv errorMsg
         return "fail";
      }
      
      def boInfo = BusinessObjectTypeInfo.getByBusinessObjectTypeName(boTypeName);
      if (boInfo == null) {
         errorMsg = "Cannot retrieve businessObjectTypeName - boInfo is null from ${boType}/${boRk}";
         plv errorMsg
         return "fail";
      }
      
      obj = boInfo.getClassObject().fetchIfExists(boRk, ps);

      if (obj == null) {
         plv "Obj is null: The specified object (boRk=${boRk}) has been deleted and no longer exists in the live table ${schema}${busObjNm}_L.", 1
         plv "-----validate objRk value (${boRk}) from the history table ${schema}${busObjNm}.", 1
         
         def custObjRk = readValue "select ${boType}_RK as custObjRk from ${schema}${boType} where ${boType}_rk = ${boRk}", "custObjRk"
         if (custObjRk == null) {
            errorMsg = "Invalid boRk=${boRk} is specified from ${boType}/${boRk}";
            plv errorMsg
            return "fail";
         }
         
         plv "-----objRk value (${objRk}) is valid by reading the record ${objRk} from the history table ${schema}${busObjNm}.", 2
         
      }
      
   } else {
      
      //-----------------------------------
      // From PurgeObjects data unloader
      //-----------------------------------
            
      obj = binding.hasVariable('businessObject') ? businessObject : null;
      
      if (obj == null) {         
         // Need to check whether it's coming from the case of the deleted object which only exists from the history table
         def ssc = binding.hasVariable('sourceSystemCode') ? sourceSystemCode : null;
         def objId = binding.hasVariable('busObjId') ? busObjId : null;
         
         plv "Obj is null: The specified object (${ssc}/${objId}) has been deleted and no longer exists in the live table ${schema}${busObjNm}_L.", 1
         plv "-----Need to read objRk value per ${ssc}/${objId} from the history table ${schema}${busObjNm}.", 1
         
         if (ssc != null && objId != null) {
            objRk = readValue "select ${busObjNm}_RK as custObjRk from ${schema}${busObjNm} where source_system_cd = '${ssc}' and ${busObjNm}_id = '${objId}'", "custObjRk"
         }

         plv "-----objRk is ${objRk} by reading the record ${ssc}/${objId} from the history table ${schema}${busObjNm}.", 2
                  
         if (objRk == null) {
            errorMsg = "Cannot find the record ${ssc}/${objId} from the live table ${schema}${busObjNm}_L or the history table ${schema}${busObjNm}";
            plv errorMsg
            return "fail";
         }
      }
   
   }

   
   def inTestMode = binding.hasVariable('inTestMode') ? inTestMode : true;
   
   plv "====================================================================================================="
   plv "Purging objects for '${obj}' : objectRk=${objRk}, objectType=${busObjNm} (testMode=${inTestMode})"
   plv "=====================================================================================================", 1
   
   if (obj == null && objRk == null) {
      errorMsg = "obj is null from ${busObjNm}/${objRk}";
      plv errorMsg      
      return "fail";
   }
   
   //----------------------------------------------------------------------------
   // delete the workflow instance
   //----------------------------------------------------------------------------
   if (!inTestMode && busObjNm != null && objRk != null) {
      def tpHandle = new TaggedPointerHandle(busObjNm, objRk);
      plv "-----Delete the workflow instance ${tpHandle}.", 2
      WorkflowUtil.deleteWorkflowInstance(tpHandle);
   }
   
   //----------------------------------------------------------------------------
   // fetch the list of the attachmentRk before deleting the custom object.
   //----------------------------------------------------------------------------
   def attachmentRks = getAttachmentRks(obj, busObjNm, objRk);
   plv "-----a list of attachmentRks : ${attachmentRks}", 2
   
   
   //----------------------------------------------------------------------------
   // fetch the list of the linkInstanceRk before deleting the custom object.
   //----------------------------------------------------------------------------
   def linkInstanceRks = getLinkInstanceRks(obj, boTypeName, objRk);
   plv "-----a list of linkInstanceRks : ${linkInstanceRks}", 2
   
   
   //------------------------------------------------------------------------------
   // delete the main custom object and its dependent objects from the live table.
   //------------------------------------------------------------------------------
   if (obj != null) {
      plv "-----Deleting ${objRk} from the live table ${schema}${busObjNm}_L", 2
      obj.delete("purging objects", ps)   
   }

   
      
   /*
    *************************************************************************************
    **
    **  SQL statements
    **
    *************************************************************************************
    */

   // the size of the bucket when splitting a list of Rks to multiple buckets.
   def bucketSize = 100; // this size can be adjusted to the other value (for instance: 30).
   
   //----------------------------------------------------------------------------
   // delete the link_instance history table.
   //----------------------------------------------------------------------------
   plv "-----Deleting the records from ${schema}LINK_INSTANCE table (bucketSize=${bucketSize})", 1
   def linkInstanceBuckets = splitToBuckets(linkInstanceRks, bucketSize);      
   linkInstanceBuckets.each {linkInstanceBucket ->
      def liBucketSql = "delete from ${schema}LINK_INSTANCE where LINK_INSTANCE_RK in (${linkInstanceBucket.join(',')})";
      execSqlParams liBucketSql;
   }
   plv "", 1
   
   
   //----------------------------------------------------------------------------
   // delete the dependent objects from the history table.
   //----------------------------------------------------------------------------
   def historyTables = ["CHANGE_REASON", "DIM_POINT_MAPPING", "ATTACHMENT", "GENERIC_COMMENT", "CUST_FIELD_VALUE"]
   def numTables = historyTables.size();
   for (def i = 0; i < numTables; i++) {
      def table = historyTables[i];
      plv "-----Deleting the records from ${schema}${table} table", 1
      execSqlParams "delete from ${schema}${table} where business_object_rk = ? and business_object_nm = ?", [objRk, busObjNm], 2;
   }
   
   
   //----------------------------------------------------------------------------
   // delete attachment's custom field value from CUST_FIELD_VALUE history table.
   //----------------------------------------------------------------------------
   if (attachmentRks.size() > 0) {      
      def attBuckets = splitToBuckets(attachmentRks, bucketSize);            
      plv "-----Deleting attachment's custom field value from ${schema}CUST_FIELD_VALUE table", 1
      attBuckets.each {attBucket ->
         def bucketSql = "delete from ${schema}CUST_FIELD_VALUE where BUSINESS_OBJECT_NM = 'ATTACHMENT' and BUSINESS_OBJECT_RK in (${attBucket.join(',')})";
         execSqlParams bucketSql;         
      }      
      plv "", 1
   }
   
   
   //------------------------------------------------------------------------------
   // delete linkInstance's custom field value from CUST_FIELD_VALUE history table.
   //------------------------------------------------------------------------------
   if (linkInstanceRks.size() > 0) {      
      plv "-----Deleting linkInstance's custom field value from ${schema}CUST_FIELD_VALUE table", 1            
      linkInstanceBuckets.each {linkInstanceBucket ->
         def liBucketSql = "delete from ${schema}CUST_FIELD_VALUE where BUSINESS_OBJECT_NM = 'LINK_INSTANCE' and BUSINESS_OBJECT_RK in (${linkInstanceBucket.join(',')})";
         execSqlParams liBucketSql;
      }
   
      plv "", 1
   }

   
   //-----------------------------------------------------------------------------------------------------------------
   // update createdFrom to null from CUST_OBJ_# history table for those objects created from the deleted object.
   //-----------------------------------------------------------------------------------------------------------------
   plv "-----Updating createdFrom to null from ${schema}${busObjNm} table for those objects created from ${objRk}", 1
   execSqlParams "update ${schema}${busObjNm} set CREATED_FROM_RK = null where CREATED_FROM_RK = ?", [objRk], 2;
   
   
   //----------------------------------------------------------------------------
   // delete from the main custom object history table.
   //----------------------------------------------------------------------------
   plv "-----Delete ${objRk} from ${schema}${busObjNm} table", 1
   execSqlParams "delete from ${schema}${busObjNm} where ${busObjNm}_rk = ?", [objRk], 2;

      
      
   /*
    *************************************************************************************
    **
    **   Handle "already-deleted" custom object - objRk only exists in the history table.
    **      delete attachment files from the content server.
    **      delete the finished workflow instance from the workflow system.
    **
    *************************************************************************************
    */
   
   if (obj == null && objRk != null && !inTestMode) {
      plv "-----Delete attachments from ${boTypeName}/${objRk} folder in the content server.", 2
      BusinessObjectUtils.permanentlyDeleteAllContent(boTypeName, objRk, ps);
   }   
   
   return "okay";   
   
} catch (Exception e) {
   
   logger.error(e.getMessage())
   println "Error occurred : ${e.getMessage().toString()}"
   e.printStackTrace();
   
   ps.rollback();
   
   errorMsg = e.getMessage();
   return "fail";
   
}
