/**
 * SAS extension of sap.m.MessageStrip for HTML Commons.
 * Copyright 2016 SAS Institute Inc.
 * All Rights Reserved.
 * @name sas.hc.m.MessageStrip
 * @extends sap.m.MessageStrip
*/

sap.ui.define([
    "sap/m/MessageStrip",
    "sap/ui/core/Control",
    "./Text"
], function(SapMessageStrip, Control, Text) {
    "use strict";

    /**
     * Constructor for a new MessageStrip.
     *
     * @param {string} [sId] id for the new control, generated automatically if no id is given
     * @param {object} [mSettings] initial settings for the new control
     *
     * @class
     * MessageStrip is a control that enables the embedding of application-related messages in the application. There are 4 types of messages: Information, Success, Warning and Error. Each message can have a close button, so that it can be removed from the UI if needed.
     *
     * @extends sap.m.MessageStrip
     * @version 800001.10.0.20241202143000_f0htmcm80p
     *
     * @constructor
     * @public
     * @alias sas.hc.m.MessageStrip
     */
    var MessageStrip = SapMessageStrip.extend("sas.hc.m.MessageStrip", {
        metadata : {
            library : "sas.hc.m",

            properties: {
                "messageIcon": {type: "object", defaultValue: undefined},

                /**
                 * Indicates that the strip has a transparent background when true. When false the background
                 * is based on the type property.
                 */
                "transparentBackground": { type: "boolean", defaultValue: false },

                /**
                 * Based on UX requirements the close button is required in the SAS version of the control.
                 */
                showCloseButton: { type: "boolean", group: "Appearance", defaultValue: true }
            },

            aggregations: {
                "successIcon": {type: "sap.m.Image", multiple: false, visibility: "hidden"},
                "errorIcon": {type: "sap.m.Image", multiple: false, visibility: "hidden"},
                "warningIcon": {type: "sap.m.Image", multiple: false, visibility: "hidden"},
                "infoIcon": {type: "sap.m.Image", multiple: false, visibility: "hidden"},
                "closeButton": { type: "sas.hc.m.Button", multiple: false},
                "helpPopup": { type: "sas.hc.m.HelpPopup", multiple: false }
            }
        },

        renderer : "sas.hc.m.MessageStripRenderer"
    });

    /**
     * Based on UX requirements the close button is required in the SAS version of the control.
     * Attempting to set the value to anything other than true will be ignored, and fetching the
     * value will always return true, regardless of any internal state of the property.
     * @param bShowButton Whether or not to show the close button. IGNORED.
     * @returns {sas.hc.m.MessageStrip} this pointer to allow method chaining
     */
    MessageStrip.prototype.setShowCloseButton = function(bShowButton) {
        this.setProperty("showCloseButton", true);
        return this;
    };

    /**
     * Based on UX requirements the close button is required in the SAS version of the control.
     * This function always returns TRUE, regardless of attempts to disable the close button.
     * @returns {boolean} Flag indicating whether to show the close button. ALWAYS TRUE.
     */
    MessageStrip.prototype.getShowCloseButton = function() {
        return true;
    };

    MessageStrip.prototype.init = function() {
        SapMessageStrip.prototype.init.apply(this, arguments);

        var self = this;
        var rb = sap.ui.getCore().getLibraryResourceBundle("sas.hc.m");
        var oCloseButton = new sas.hc.m.Button({
            icon: sas.icons.HC.CLOSE,
            press: function() {
                self.close();
            }
        });

        // Add both SAP and SAS style classes.
        oCloseButton.addStyleClass("sapMMsgStripCloseButton");
        oCloseButton.addStyleClass("sasMMsgStripCloseButton");
        oCloseButton.setAriaLabel(rb.getText("messageStrip.closeButton.label.txt"));

        this.setCloseButton(oCloseButton);
    };

    MessageStrip.prototype._createText = function() {
        var oText = new Text();
        return oText;
    };

    MessageStrip.prototype.setHelpPopup = function(oHelpPopup) {
        oHelpPopup.addStyleClass("sasMMsgStripHelp");
        this.setAggregation("helpPopup", oHelpPopup);
    };

    MessageStrip.prototype.setAggregation = function (sName, oControl, bSupressInvalidate) {
        // SAP's MessageStrip forces the _text aggregation to include the MessageStrip as
        // aria-labelledby of the Text. For our own improved screen reader UX, we don't want that.
        return Control.prototype.setAggregation.call(this, sName, oControl, bSupressInvalidate);
    };

    /**
     * S1281317
     * Fix memory leaks with rendering of icons
     *
     * HTMLCOMMONS-6605
     * Create image tag object for icon.  This method is intended for
     * internal use only.  It is called by the getIcon() method.
     *
     * @param iconSrc name of icon
     * @param iconDecorative indicate if decorative or not
     * @param iconAltText alt text for image
     * @returns {sas.hc.m.Image} image object using icon information passed to method
     * @private
     */
    MessageStrip.prototype._createIcon = function(iconSrc, iconDecorative, iconAltText) {

        //returns an image
        return sap.ui.core.IconPool.createControlByURI({
            id : this.getId() + "-icon",
            src : iconSrc,
            decorative : iconDecorative,
            alt : iconAltText
        });
    };

    /**
     * Return icon associated with the current control type
     *
     * @returns {sas.hc.m.Image} Image instance that represents the current control Message type.  The types are complete, error,
     * warning and information
     */
    MessageStrip.prototype.getIcon = function() {
        var type = this.getType(),  //has default value of information
            rb = sap.ui.getCore().getLibraryResourceBundle("sas.hc.m"),
            _successIcon=null, _errorIcon=null, _warningIcon=null, _infoIcon=null, _messageIcon = this.getMessageIcon();

        // check if image source has changed - if yes destroy and reset image control
        //since there is a setType, we need to make sure current icon reflects the correct type
        if (_messageIcon && !this._iconMatchesMessageType(type, _messageIcon)) {
            _messageIcon.destroy();
            _messageIcon = null;
        }

        if(!_messageIcon) {
            if (type === sap.ui.core.MessageType.Success) {
                _successIcon = this.getAggregation("successIcon");
                if (!_successIcon) {
                    _successIcon = this._createIcon(sas.icons.HC.COMPLETESTATUS, false, rb.getText("messageStrip.message.success.txt"));
                    this.setAggregation("successIcon", _successIcon);
                }
                this.setMessageIcon(_successIcon);
                return _successIcon;
            }
            else if (type === sap.ui.core.MessageType.Error) {
                _errorIcon = this.getAggregation("errorIcon");
                if (!_errorIcon) {
                    _errorIcon = this._createIcon(sas.icons.HC.ERRORSTATUS, false, rb.getText("messageStrip.message.error.txt"));
                    this.setAggregation("errorIcon", _errorIcon);
                }
                this.setMessageIcon(_errorIcon);
                return _errorIcon;
            }
            else if (type === sap.ui.core.MessageType.Warning) {
                _warningIcon = this.getAggregation("warningIcon");
                if (!_warningIcon) {
                    _warningIcon = this._createIcon(sas.icons.HC.WARNINGSTATUS, false, rb.getText("messageStrip.message.warning.txt"));
                    this.setAggregation("warningIcon", _warningIcon);
                }
                this.setMessageIcon(_warningIcon);
                return _warningIcon;
            }
            else { //return default type of Information
                _infoIcon = this.getAggregation("infoIcon");
                if (!_infoIcon) {
                    _infoIcon = this._createIcon(sas.icons.HC.INFORMATIONSTATUS, false, rb.getText("messageStrip.message.information.txt"));
                    this.setAggregation("infoIcon", _infoIcon);
                }
                this.setMessageIcon(_infoIcon);
                return _infoIcon;
            }
        }
    };

    MessageStrip.prototype._iconMatchesMessageType = function(type, _messageIcon) {

        return ( (type===sap.ui.core.MessageType.Success && _messageIcon.src === sas.icons.HC.COMPLETESTATUS) ||
             (type === sap.ui.core.MessageType.Error && _messageIcon.src === sas.icons.HC.ERRORSTATUS) ||
            (type === sap.ui.core.MessageType.Warning && _messageIcon.src === sas.icons.HC.WARNINGSTATUS) ||
            (_messageIcon.src === sas.icons.HC.INFORMATIONSTATUS) );
    };

    MessageStrip.prototype.exit = function() {
        var _messageIcon = this.getMessageIcon();

        if (_messageIcon) {
            _messageIcon.destroy();
            _messageIcon = null;
        }

        if (SapMessageStrip.prototype.exit) {
            SapMessageStrip.prototype.exit.apply(this, arguments);
        }
    };
    return MessageStrip;
});

