/*!
* (c) 2014-2018, SAS Institute Inc.
*/

// Provides control sas.hc.ui.table.Column.
sap.ui.define([
    'jquery.sap.global',
    'sas/hc/ui/core/AnimationUtil',
    'sap/ui/core/Element',
    'sap/ui/core/library',
    'sap/ui/core/Popup',
    'sap/ui/core/RenderManager',
    'sap/ui/model/Filter',
    'sap/ui/model/FilterOperator',
    'sap/ui/model/FilterType',
    'sap/ui/model/Sorter',
    'sap/ui/model/Type',
    'sap/ui/model/type/String',
    'sas/hc/ui/table/library',
    './TableUtils'
],
function(jQuery, AnimationUtil, Element, coreLibrary, Popup, RenderManager, Filter, FilterOperator, FilterType, Sorter, Type, StringType, library, TableUtils) {
    "use strict";

    // shortcuts
    var HorizontalAlign = coreLibrary.HorizontalAlign,
        SortOrder = library.SortOrder,
        SharedDomRef = library.SharedDomRef,
        ValueState = coreLibrary.ValueState;

    // var logger = library._getLogger("Column");

    /**
     * Constructor for a new Column.
     *
     * @param {string} [sId] ID for the new control, generated automatically if no ID is given
     * @param {object} [mSettings] Initial settings for the new control
     *
     * @class
     * The column allows you to define column specific properties that will be applied when rendering the table.
     * @extends sap.ui.core.Element
     * @version 800001.10.0.20241202143000_f0htmcm80p
     *
     * @constructor
     * @public
     * @alias sas.hc.ui.table.Column
     * @ui5-metamodel This control/element also will be described in the UI5 (legacy) designtime metamodel
     */
    var Column = Element.extend("sas.hc.ui.table.Column", /** @lends sas.hc.ui.table.Column.prototype */ { metadata : {

        library : "sas.hc.ui.table",
        properties : {

            /**
             * Width of the column. Works only with px/em/rem values. Em are handled like rem values.
             */
            width : {type : "sap.ui.core.CSSSize", group : "Dimension", defaultValue : null},

            /**
             * If the table is wider than the sum of widths of the visible columns, the columns will be
             * resized proportionally to their widths that were set originally. If set to false, the column will be displayed in the
             * original width. If all columns are set to not be flexible, an extra "dummy" column will be
             * created at the end of the table.
             */
            flexible : {type : "boolean", group : "Behavior", defaultValue : true},

            /**
             * If set to true, the column can be resized either using the resize-handle (by mouse) or using
             * the keyboard (SHIFT + Left/Right Arrow keys)
             */
            resizable : {type : "boolean", group : "Behavior", defaultValue : true},

            /**
             * Horizontal alignment of the column content. Controls with a text align do not inherit
             * the horizontal alignment. You have to set the text align directly on the template.
             */
            hAlign : {type : "sap.ui.core.HorizontalAlign", group : "Appearance", defaultValue : HorizontalAlign.Begin},

            /**
             * Indicates if the column is sorted. This property only controls if a sort indicator is displayed in the
             * column header - it does not trigger the sort function. The column has to be sorted by calling <code>Column.sort()</code>
             */
            sorted : {type : "boolean", group : "Appearance", defaultValue : false},

            /**
             * This property indicates the sort direction (Ascending or Descending). The corresponding icon will be
             * rendered if the property <code>sorted</code> is <code>true</code>
             * @see sas.hc.ui.table.SortOrder (default value: "Ascending")
             */
            sortOrder : {type : "sas.hc.ui.table.SortOrder", group : "Appearance", defaultValue : SortOrder.Ascending},

            /**
             * Specifies the binding property on which the column will sort.
             * Since the column template may have composite bindings, it's not possible to figure out on which binding
             * property the sort shall be applied. Therefore the binding property for sorting must be specified.
             * For example, if the first name and last name are displayed in the same column, only one of the two can be defined as
             * <code>sortProperty</code>.
             *
             * A column menu entry for sorting can only be generated if the <code>sortProperty</code> is set.
             */
            sortProperty : {type : "string", group : "Behavior", defaultValue : null},

            /**
             * Indicates if the column is filtered. This property only controls if a filter indicator is displayed in the
             * column header - it does not trigger the filter function. The column has to be filtered by calling <code>Column.sort()</code>
             */
            filtered : {type : "boolean", group : "Appearance", defaultValue : false},

            /**
             * Specifies the binding property on which the column shall be filtered.
             * Since the column template may have composite bindings, it's not possible to figure out on which binding
             * property the filter shall be applied. Therefore the binding property for filtering must be specified.
             * For example, if the first name and last name are displayed in the same column, only one of the two can be defined as
             * <code>filterProperty</code>.
             *
             * A column menu entry for filtering can only be generated if the <code>filterProperty</code> is set. The
             * default menu entry is a text input field.
             */
            filterProperty : {type : "string", group : "Behavior", defaultValue : null},

            /**
             * Specifies the value of the filter as string (will be converted into the proper data type). It is possible
             * to provide a filterOperator as string, as shown here:
             * <pre>
             * &gt; 50
             * &lt; 100
             * &gt;= 150
             * &lt;= 200
             * = 250
             * != 300
             * *something    ends with
             * something*    starts with
             * *something*   contains
             * some..thing   between
             * 50..100       between
             * </pre>
             */
            filterValue : {type : "string", group : "Behavior", defaultValue : null},

            /**
             * Filter operator to use when filtering this column.
             * @see sap.ui.model.FilterOperator (default value: "Contains")
             */
            filterOperator : {type : "string", group : "Behavior", defaultValue : null},

            /**
             * If this property is set, the default filter operator of the column is overridden.
             * By default <code>Contains</code> is used for string and <code>EQ</code> for other types. A valid <code>sap.ui.model.FilterOperator</code> needs to be passed.
             */
            defaultFilterOperator : {type : "string", group : "Behavior", defaultValue : null},

            /**
             * Type of Filter. This is used to transform the search term to the specified type,
             * to make sure that the right columns are displayed. This should be the same as defined
             * in binding for this column. As an alternative you can pass a function which does the conversion.
             * The function receives the entered filter value as parameter and returns the proper
             * value for the filter expression. Another option is to pass the class name of the type,
             * e.g.: <code>sap.ui.model.type.Date</code> or an expression similar to the binding syntax,
             * e.g.: <code>"\{type: 'sap.ui.model.type.Date', formatOptions: \{UTC: true\}, constraints: {} \}"</code>.
             * Here the escaping is mandatory to avoid handling by the binding parser.
             * By default the filter type is <code>sap.ui.model.type.String</code>.
             * @since 1.9.2
             */
            filterType : {type : "any", group : "Misc", defaultValue : null},

            /**
             * Indicates if the column is grouped.
             */
            grouped : {type : "boolean", group : "Appearance", defaultValue : false},

            /**
             * Invisible controls are not rendered.
             */
            visible : {type : "boolean", group : "Appearance", defaultValue : true},

            /**
             * Required columns are always visible. Default value is false.
             */
            required : {type : "boolean", group : "Appearance", defaultValue : false},

            /**
             * Mandatory columns are required columns that also cannot be moved and are always frozen.
             */
            "mandatory": {type: "boolean", group: "Behavior", defaultValue: false},

            /**
             * The name of the column which is used in the column visibility menu item as text.
             * If not set as a fallback the column menu tries to get the text from the nested Label.
             * @since 1.11.1
             */
            name : {type : "string", group : "Appearance", defaultValue : null},

            /**
             * Defines if the filter menu entry is displayed
             * @since 1.13.0
             */
            showFilterMenuEntry : {type : "boolean", group : "Appearance", defaultValue : true},

            /**
             * Defines if the sort menu entries are displayed
             * @since 1.13.0
             */
            showSortMenuEntry : {type : "boolean", group : "Appearance", defaultValue : true},

            /**
             * If this property is set, a span is applied for the header. When moving columns, all columns
             * which are part of the header will be moved. The <code>headerSpan</code> can be either an integer or an array of
             * integers (if you use the multi header feature of the table). If you only specify an integer, this span is
             * applied for all header rows, with multiple integers you can specify a separate span for each header row.
             */
            headerSpan : {type : "any", group : "Behavior", defaultValue : 1},

            /**
             * Enables auto-resizing of the column on double-clicking the resizer. The width is determined on the widest
             * currently displayed content. It does not consider rows which are currently not scrolled into view.
             * Currently only implemented to work with the following controls:
             * <code>sap.m.Text, sap.m.Label, sap.m.Link, sap.m.Input,
             * sap.ui.commons.Label, sap.ui.commons.Link,
             * sap.ui.commons.Checkbox, sap.m.Checkbox</code>
             * @since 1.21.1
             */
            autoResizable : {type : "boolean", group : "Behavior", defaultValue : false},

            "highlighted": {type: "boolean", defaultValue: false},
            "styleClass": {type: "string", group: "Appearance", defaultValue: null},
            "sortFunction": {type: "object", group: "Behavior"},
            "headerEditable": {type: "boolean", group: "Behavior", defaultValue: false},

            /**
             * Passed to sap.ui.model.Sorter objects that are created from this column
             *
             * Note:
             * This property could be deceiving if a collator service is not used!
             * JavaScript's "localeCompare" function cannot properly sort by case, so
             * if a collator is not used this property will seem to not be working.
             * More information: http://stackoverflow.com/questions/34861386/force-localecompare-to-be-case-sensitive
             */
            "caseSensitiveSort": {type: "boolean", group: "Behavior", defaultValue: false}
        },
        defaultAggregation : "label",
        aggregations : {

            /**
             * Label of the column which is displayed in the column header. This aggregation is for the standard behavior,
             * where you only want to display one single row header. If a string is supplied, a default label control will be
             * created. Which control this is depends on the loaded libraries.
             */
            label : {type : "sap.ui.core.Control", altTypes : ["string"], multiple : false},

            /**
             * Labels of the column which are displayed in the column header. Define a control for
             * each header row in the table. Use this aggregation if you want to use multiple headers per column.
             * @since 1.13.1
             */
            multiLabels : {type : "sap.ui.core.Control", multiple : true, singularName : "multiLabel"},

            /**
             * Template (cell renderer) of this column. A template is decoupled from the column. Each time
             * the template's properties or aggregations have been changed, the template has to be applied again via
             * <code>setTemplate</code> for the changes to take effect. The default template depends on the libraries loaded.
             */
            template : {type : "sap.ui.core.Control", multiple : false},

            /**
             * The menu used by the column. By default the {@link sas.hc.ui.table.ColumnMenu} is used.
             *
             * <b>Note:</b> Applications must not use or change the default <code>sas.hc.ui.table.ColumnMenu</code> of
             * a column in any way or create own instances of <code>sas.hc.ui.table.ColumnMenu</code>.
             * To add a custom menu to a column, use the aggregation <code>menu</code> with a new instance of
             * <code>sap.ui.unified.Menu</code>.
             */
            menu : {type : "sap.ui.unified.Menu", multiple : false},

            "cells": {type: "sap.ui.core.Control", multiple: true, singularName: "cell", bindable: "bindable"},
            /**
              * summaryLabel (header renderer) of the column which is displayed in the column header.
              */
            summaryLabel : {type : "sap.ui.core.Control", multiple : false},
            /**
             * summaryControl (header control) of the column which is displayed in the column summary header.
             */
            summaryControl: {type : "sap.ui.core.Control", multiple : false}
        },

        events : {
            /**
             * Fires before the column menu is opened.
             * @since 1.33.0
             */
            columnMenuOpen: {
                allowPreventDefault: true,
                parameters: {
                    /**
                     * Refence to the selected <code>menu</code> instance to be opened.
                     */
                    menu: {type: "sap.ui.unified.Menu"}
                }
            }
        }
    }});


    /** default filter type for the columns */
    Column._DEFAULT_FILTER_TYPE = new StringType();

    /**
     * called when the column is initialized
     */
    Column.prototype.init = function() {

        this.oResBundle = sap.ui.getCore().getLibraryResourceBundle("sas.hc.ui.table");
        this._oSorter = null;

        // used to track if we are in keyboard resizing mode
        this._bKeyboardResizing = false;

        // Skip proppagation of databinding properties to the template
        this.mSkipPropagation = {template: true};

        this._oInPlaceEdit = new sas.hc.ui.commons.InPlaceEdit({
            undoEnabled: false,
            inputProperties: {
                required: true
            },
            revertToCommittedValue: true
        }).addStyleClass("sasUiTableColHdrIpe");
        var that = this;
        this._oInPlaceEdit.getContent().attachEvent('change', function(oEvent){// Fix S1231817
            var sValue = oEvent.getParameter("newValue");

            // For some reason we get an error from the rerender
            // function when user clears out the header. So just
            // abort this update if the content value is empty.
            if (jQuery.trim(sValue) === "") {
                return;
            }

            if (that._oInPlaceEdit._oTextView) that._oInPlaceEdit._oTextView.setText(sValue);
            that.setIpeValue(sValue);
            var oTable = that.getParent();
            if (oTable) oTable.invalidate();
        });
    };

    /**
     * cleans up when exiting
     *
     * @returns {Object} column
     */
    Column.prototype.exit = function() {
        var oTable = this.getParent();
        this._oInPlaceEdit.destroy();
        jQuery(window).off("keydown.tbl_keyboard_resize");
        this._getKeyboardResizeReference().off("blur.tbl_keyboard_resize");
        if (oTable) {
            jQuery(oTable.getDomRef(SharedDomRef.HorizontalScrollBar)).off("mousedown.tbl_keyboard_resize");
            jQuery(oTable.getDomRef(SharedDomRef.VerticalScrollBar)).off("mousedown.tbl_keyboard_resize");
        }
    };

    // Purposfully no JSDoc to use the auto-generated version.
    Column.prototype.setWidth = function(sWidth) {
        if (!this.hasOwnProperty("_sConfiguredWidth")) {
            this._sConfiguredWidth = sWidth || "";
        }
        return this.setProperty("width", sWidth);
    };

    /**
     * @override
     */
    Column.prototype.setParent = function(oParent, sAggregationName, bSuppressRerendering) {
        Element.prototype.setParent.apply(this, arguments);
        var oMenu = this.getAggregation("menu");
        if (oMenu && typeof oMenu._updateReferences === "function") {
            //if menu is set update menus internal references
            oMenu._updateReferences(this);
        }
    };

    /**
     * Determine the reference element to use for the keyboard resize operation.
     * If summary headers are in use, use those (since the col header may be hidden),
     * otherwise use the column header itself
     * @returns {object} jQuery reference
     * @private
     */
    Column.prototype._getKeyboardResizeReference = function() {
        var oTable = this.getParent();
        return (oTable && oTable.getShowSummaryHeader() === true) ?
            oTable.$().find('.sasUiTableColSumHead').eq(this.getIndex()) :
            this.$();
    };

    /**
     * Manage the _bKeyboardResizing instance variable
     * Add or remove the css class that shows a visual indication of keyboard resizing
     * @param {boolean} bResizing if we are in keyboard resizing mode or not
     * @private
     */
    Column.prototype._startKeyboardResizing = function(bResizing) {
        var $colRef = this._getKeyboardResizeReference();
        this._bKeyboardResizing = bResizing;
        if (bResizing === true) {
            $colRef.addClass("sapUiTableColKeyboardResizeInProcess");
        } else {
            $colRef.removeClass("sapUiTableColKeyboardResizeInProcess");
        }
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.invalidate = function(oOrigin) {
        // prevent changes in the template (especially the databinding ones)
        //  - what about exchanging the template? => implemented in setTemplate
        //  - what about modifiying properties? => developer must call invalidate!
        // The problem is that we just need to prevent databinding changes. The
        // problem here is that the databinding bindings are created ones the template
        // is created and has its own model. If now changes are done in the model
        // this directly affects the template which invalidates the column invalidating
        // the complete Table.
        /*
         * PART1: When you create the Tooltip (deferred) then it establishes the
         * connection to its data (also for the template of the column!) and this
         * finally invalidates the Table which triggers the re-rendering. One
         * option is to complete decouple the template from the Table by
         * supressing the invalidate. But this finally also decouples the Table
         * from any changes on the template after the template has been applied
         * to the Column. But when re-rendering it would update the column cells.
         * To notify the Table on proper changes one has to call the method
         * invalidate on the Table.
         */
        /*
         * PART2: we also suppress the re-rendering in case of the column menu is
         * rerendered. This is a popup and we use the instance check because of the
         * menu behind the getMenu function is lazy created when first accessed.
         */
        var ColumnMenu = sap.ui.require("sas/hc/ui/table/ColumnMenu");
        if (oOrigin !== this.getTemplate() && !(ColumnMenu && oOrigin instanceof ColumnMenu)) {
            // changes on the template require to call invalidate on the column or table
            Element.prototype.invalidate.apply(this, arguments);
        }
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setLabel = function(vLabel) {
        var oLabel = vLabel;
        var bHeaderEditable = this.getHeaderEditable();
        var oPreviousLabel = this.getLabel();

        if (oPreviousLabel === oLabel) {
            return;
        }

        if (typeof (vLabel) === "string" && bHeaderEditable) {
            this.setIpeValue(vLabel);
            oLabel = this._oInPlaceEdit;
        } else if (typeof (vLabel) === "string") {
            oLabel = library.TableHelper.createLabel({text: vLabel});
        }

        this.setAggregation("label", oLabel);

        // S1409768 - Memory leak
        // The label aggregation can be a Control or a string
        // If a new Control is passed in or created, destroy the old one
        // unless it is the internal _oInPlaceEdit
        if (oPreviousLabel && oPreviousLabel !== this._oInPlaceEdit) {
            oPreviousLabel.destroy();
        }

        return this;
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setTemplate = function(vTemplate) {
        var oTemplate = vTemplate;
        if (typeof (vTemplate) === "string") {
            oTemplate = library.TableHelper.createTextView().bindProperty("text", vTemplate);
        }
        this.setAggregation("template", oTemplate);
        // manually invalidate the Column (because of the invalidate decoupling to
        // prevent invalidations from the databinding part)
        this.invalidate();
        var oTable = this.getParent();
        if (oTable && oTable._resetRowTemplate) {
            oTable._resetRowTemplate();
        }
        return this;
    };


    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.getMenu = function() {
        var oMenu = this.getAggregation("menu");
        if (!oMenu) {
            oMenu = this._createMenu();
            this.setMenu(oMenu);
        }
        return oMenu;
    };

    /**
     * This function invalidates the column's menu. All items will be re-created the next time the menu opens. This only
     * happens for generated menus.
     * @private
     */
    Column.prototype.invalidateMenu = function() {
        var oMenu = this.getAggregation("menu");
        if (oMenu && oMenu._invalidate) {
            oMenu._invalidate();
        }
    };

    /**
     * Checks whether or not the menu has items. This function considers table and column
     * properties to determine whether the column menu would have items. If there is a menu set,
     * it will just check whether there are items in the item aggregation.
     * @return {Boolean} True if the menu has or could have items.
     * @private
     */
    Column.prototype._menuHasItems = function() {
        var oMenu = this.getAggregation("menu");
        var oTable = this.getParent();
        var fnMenuHasItems = function() {
            return (
                this.isSortableByMenu() || // Sorter
                this.isFilterableByMenu() || // Filter
                this.isGroupableByMenu() || // Grouping
                (oTable && oTable.getEnableColumnFreeze()) || // Column Freeze
                (oTable && oTable.getShowColumnVisibilityMenu()) // Column Visibility Menu
            );

        }.bind(this);

        return !!((oMenu && oMenu.getItems().length > 0) || fnMenuHasItems());
    };

    /**
     * This function checks whether a filter column menu item will be created. Although it evaluates some column
     * properties, it does not check the metadata.
     *
     * For Columns the following applies:
     * - filterProperty must be defined
     * - showFilterMenuEntry must be true (which is the default)
     *
     * @public
     * @returns {boolean}
     */
    Column.prototype.isFilterableByMenu = function() {
        return !!(this.getFilterProperty() && this.getShowFilterMenuEntry());
    };

    /**
     * This function checks whether sort column menu items will be created. Although it evaluates some column
     * properties, it does not check the metadata.
     *
     * For Columns the following applies:
     * - sortProperty must be defined
     * - showSortMenuEntry must be true (which is the default)
     *
     * @public
     * @returns {boolean}
     */
    Column.prototype.isSortableByMenu = function() {
        return !!(this.getSortProperty() && this.getShowSortMenuEntry());
    };

    /**
     * This function checks whether a grouping column menu item will be created. Although it evaluates some column
     * properties, it does not check the metadata. Since a property of the table must be checked, this function will
     * return false when the column is not a child of a table.
     *
     * For Columns the following applies:
     * - sortProperty must be defined
     * - showFilterMenuEntry must be true (which is the default)
     *
     * @public
     * @returns {boolean}
     */
    Column.prototype.isGroupableByMenu = function() {
        var oTable = this.getParent();
        return !!(oTable && oTable.getEnableGrouping && oTable.getEnableGrouping() && this.getSortProperty());
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setMenu = function(oMenu) {
        this.setAggregation("menu", oMenu, true);
        return this;
    };

    /**
     * Factory method. Creates the column menu.
     *
     * @private
     * @return {sas.hc.ui.table.ColumnMenu} The created column menu.
     */
    Column.prototype._createMenu = function() {
        jQuery.sap.require("sas.hc.ui.table.ColumnMenu");
        return new sas.hc.ui.table.ColumnMenu(this.getId() + "-menu");
    };

    /**
     * @private
     */
    Column.prototype._setAppDefault = function(sProperty, mValue) {
        if (!this._appDefaults) {
            this._appDefaults = {};
        }

        if (sProperty === "sorted") {
            this._appDefaults.sorted = mValue;
        } else if (sProperty === "sortOrder") {
            this._appDefaults.sortOrder = mValue;
        } else if (sProperty === "filtered") {
            this._appDefaults.filtered = mValue;
        } else if (sProperty === "filterValue") {
            this._appDefaults.filterValue = mValue;
        } else if (sProperty === "filterOperator") {
            this._appDefaults.filterOperator = mValue;
        }
    };

    /**
     * @private
     */
    Column.prototype._restoreAppDefaults = function() {
        if (this._appDefaults) {
            this.setProperty("sorted", this._appDefaults.sorted, true);
            this.setProperty("sortOrder", this._appDefaults.sortOrder, true);
            this.setProperty("filtered", this._appDefaults.filtered, true);
            this.setProperty("filterValue", this._appDefaults.filterValue, true);
            this.setProperty("filterOperator", this._appDefaults.filterOperator, true);
            this._updateIcons();
        }
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setSortProperty = function(sValue) {
        this.setProperty("sortProperty", sValue);
        this.invalidateMenu();
        return this;
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setSorted = function(bFlag) {
        this.setProperty("sorted", bFlag, true);
        this._setAppDefault("sorted", bFlag);
        this._updateIcons();
        return this;
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setSortOrder = function(tSortOrder) {
        this.setProperty("sortOrder", tSortOrder, true);
        this._setAppDefault("sortOrder", tSortOrder);
        this._updateIcons();
        return this;
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setFilterProperty = function(sValue) {
        this.invalidateMenu();
        return this.setProperty("filterProperty", sValue);
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setFiltered = function(bFlag) {
        this.setProperty("filtered", bFlag, true);
        this._setAppDefault("filtered", bFlag);
        this._updateIcons();
        return this;
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setFilterValue = function(sValue) {
        this.setProperty("filterValue", sValue, true);
        this._setAppDefault("filterValue", sValue);
        var oMenu = this.getMenu();
        var ColumnMenu = sap.ui.require("sas/hc/ui/table/ColumnMenu");
        if (oMenu && ColumnMenu && oMenu instanceof ColumnMenu) {
            oMenu._setFilterValue(sValue);
        }
        return this;
    };

    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setFilterOperator = function(sValue) {
        this._setAppDefault("filterOperator", sValue);
        return this.setProperty("filterOperator", sValue, true);
    };

    /*
     * Required columns are always visible
     */
    Column.prototype.getVisible = function() {
        return this.getRequired() || this.getProperty("visible");
    };

    /**
     * Open the column menu
     * @param {Object} [oDomRef] DOM reference of the element to which the menu should be visually attached. Fallback is the focused DOM reference.
     * @param {boolean} [bWithKeyboard=false] Indicates whether or not the first item shall be highlighted when the menu is opened.
     * @private
     */
    Column.prototype._openMenu = function(oDomRef, bWithKeyboard) {
        var oMenu = this.getMenu();
        var bExecuteDefault = this.fireColumnMenuOpen({
            menu: oMenu
        });

        if (bExecuteDefault) {
            var eDock = Popup.Dock;
            var oFocusDomRef = oDomRef;
            if (!oDomRef) {
                oDomRef = this.getDomRef();
                oFocusDomRef = this.getFocusDomRef();
            }
            oMenu.open(!!bWithKeyboard, oFocusDomRef, eDock.BeginTop, eDock.BeginBottom, oDomRef, "none none");
        }
    };


    /**
     * Toggles the sort order of the column.
     *
     * @type sas.hc.ui.table.Column
     * @public
     * @deprecated Since version 1.5.1.
     * Please use the function "sas.hc.ui.Table.prototype.sort".
     * @ui5-metamodel This method also will be described in the UI5 (legacy) designtime metamodel
     */
    Column.prototype.toggleSort = function() {
        // by default we sort ascending / only if already is sorted ascending then we toggle
        this.sort(this.getSorted() && this.getSortOrder() === SortOrder.Ascending);
    };


    /**
     * sorts the current column ascending or descending
     *
     * @param {boolean} bDescending
     *         sort order of the column (if undefined the default will be ascending)
     * @type sas.hc.ui.table.Column
     * @public
     * @deprecated Since version 1.5.1.
     * Please use the function "sas.hc.ui.Table.prototype.sort".
     * @ui5-metamodel This method also will be described in the UI5 (legacy) designtime metamodel
     */
    Column.prototype.sort = function(bDescending, bAdd) {
        var oColumn = this,
            oTable = this.getParent(),
            // the sort order type
            sNewSortOrder = bDescending ? SortOrder.Descending : SortOrder.Ascending;

        if (oTable) {
            // notify the event listeners
            var bExecuteDefault = oTable.fireSort({
                column: this,
                sortOrder: sNewSortOrder,
                columnAdded: bAdd
            });

            // if listeners have permitted default event behavior, then proceed
            if (bExecuteDefault) {
                // update properties of current column
                oColumn.setProperty("sorted", true, true);
                oColumn.setProperty("sortOrder", sNewSortOrder, true);

                // get a Promise that resolves to the new Sorter object for the column
                this._createSorterPromise().then(function(oSorter) {
                    // save the Sorter
                    oColumn._oSorter = oSorter;

                    // add current column to list of sorted columns
                    oTable.pushSortedColumn(oColumn, bAdd);

                    // get all sorted columns as they stand after the current operation
                    var aSortedCols = oTable.getSortedColumns();
                    var aColumns = oTable.getColumns();

                    // reset the sorting status of all columns which are not sorted anymore
                    for (var i = 0, l = aColumns.length; i < l; i++) {
                        if (jQuery.inArray(aColumns[i], aSortedCols) < 0) {
                            // column is not sorted anymore -> reset default and remove sorter
                            aColumns[i].setProperty("sorted", false, true);
                            aColumns[i].setProperty("sortOrder", SortOrder.Ascending, true);
                            aColumns[i]._updateIcons();
                            delete aColumns[i]._oSorter;
                        }
                    }

                    // add sorters of all sorted columns to one sorter-array and update sort icon rendering for sorted columns
                    var aSorters = [];
                    for (var i = 0, l = aSortedCols.length; i < l; i++) {
                        aSortedCols[i]._updateIcons();
                        aSorters.push(aSortedCols[i]._oSorter);
                    }

                    if (oTable.isBound("rows")) {
                        // sort the binding
                        oTable.getBinding("rows").sort(aSorters);

                        if (oColumn._afterSort) {
                            oColumn._afterSort();
                        }
                    }
                });
            }
        }

        return this;
    };

    /**
     * Updates the filter and sort icons for this column. Will render or remove as necessary.
     * @see sas.hc.ui.table.Column.prototype._updateIcons
     * @private
     */
    Column.prototype._updateIcons = function() {
        var oTable = this.getParent(),
            bSorted = this.getSorted(),
            bFiltered = this.getFiltered(),
            $self = this.$(),
            $sort, $filter;

        if (!oTable || !oTable.getDomRef()) {
            return;
        }

        $self.find(".sapUiTableColCell")
            .toggleClass("sasUiTableColSF", bSorted || bFiltered)
            .toggleClass("sasUiTableColFiltered", bFiltered)
            .toggleClass("sasUiTableColSorted", bSorted);

        $sort = $self.find(".sasSortIcon");
        $filter = $self.find(".sasFilterIcon");

        //empty out the old icons
        if ($filter.length > 0 && !bFiltered) {
            $filter.remove();
        } else if (bFiltered && $filter.length < 1) {
            this._renderFilterIcon();
        }
        if ($sort.length > 0 && !bSorted) {
            $sort.remove(); //remove if present, but unwanted
        } else if (bSorted) {   //otherwise sort (spin existing icon, if present)
            this._renderCustomSortIcon($sort);
        }

        oTable._getAccExtension().updateAriaStateOfColumn(this);
    };

    /**
     * @private
     */
    Column.prototype._renderSortIcon = function() {
        this._updateIcons();
    };

    /**
     * @private
     */
    Column.prototype._getFilter = function() {

        var oFilter,
            sPath = this.getFilterProperty(),
            sValue = this.getFilterValue(),
            sOperator = this.getFilterOperator(),
            sParsedValue,
            sSecondaryParsedValue,
            oType = this.getFilterType() || Column._DEFAULT_FILTER_TYPE,
            bIsString = oType instanceof StringType,
            aBetween;

        if (sValue) {

            // determine the operator
            if (!sOperator) {

                aBetween = sValue.match(/(.*)\s*\.\.\s*(.*)/);

                // determine the filter operator depending on the
                if (sValue.indexOf("=") === 0) {
                    sOperator = FilterOperator.EQ;
                    sParsedValue = sValue.substr(1);
                } else if (sValue.indexOf("!=") === 0) {
                    sOperator = FilterOperator.NE;
                    sParsedValue = sValue.substr(2);
                } else if (sValue.indexOf("<=") === 0) {
                    sOperator = FilterOperator.LE;
                    sParsedValue = sValue.substr(2);
                } else if (sValue.indexOf("<") === 0) {
                    sOperator = FilterOperator.LT;
                    sParsedValue = sValue.substr(1);
                } else if (sValue.indexOf(">=") === 0) {
                    sOperator = FilterOperator.GE;
                    sParsedValue = sValue.substr(2);
                } else if (sValue.indexOf(">") === 0) {
                    sOperator = FilterOperator.GT;
                    sParsedValue = sValue.substr(1);
                } else if (aBetween) {
                    if (aBetween[1] && aBetween[2]) {
                        sOperator = FilterOperator.BT;
                        sParsedValue = aBetween[1];
                        sSecondaryParsedValue = aBetween[2];
                    } else if (aBetween[1] && !aBetween[2]) {
                        sOperator = FilterOperator.GE;
                        sParsedValue = aBetween[1];
                    } else {
                        sOperator = FilterOperator.LE;
                        sParsedValue = aBetween[2];
                    }
                } else if (bIsString && sValue.indexOf("*") === 0 && sValue.lastIndexOf("*") === sValue.length - 1) {
                    sOperator = FilterOperator.Contains;
                    sParsedValue = sValue.substr(1, sValue.length - 2);
                } else if (bIsString && sValue.indexOf("*") === 0) {
                    sOperator = FilterOperator.EndsWith;
                    sParsedValue = sValue.substr(1);
                } else if (bIsString && sValue.lastIndexOf("*") === sValue.length - 1) {
                    sOperator = FilterOperator.StartsWith;
                    sParsedValue = sValue.substr(0, sValue.length - 1);
                } else {
                    if (this.getDefaultFilterOperator()) {
                        sOperator = this.getDefaultFilterOperator();
                    } else {
                        if (bIsString) {
                            // Due to compatibility reason we need to use Contains for Strings instead of EQ as default!!
                            sOperator = FilterOperator.Contains;
                        } else {
                            sOperator = FilterOperator.EQ;
                        }
                    }
                    sParsedValue = sValue.substr(0);
                }
                if (!sSecondaryParsedValue) {
                    oFilter = new Filter(sPath, sOperator, this._parseFilterValue(sParsedValue));
                } else {
                    oFilter = new Filter(sPath, sOperator, this._parseFilterValue(sParsedValue), this._parseFilterValue(sSecondaryParsedValue));
                }
            } else {
                oFilter = new Filter(sPath, sOperator, this._parseFilterValue(sValue));
            }

        }

        return oFilter;

    };

    /**
     * Filters the Column based on the provided filter value.
     * @public
     * @param {string} sValue The filter value
     * @returns {sas.hc.ui.table.Column} This instance of Column.
     */
    Column.prototype.filter = function(sValue) {
        var oTable = this.getParent();

        if (oTable && oTable.isBound("rows")) {

            // notify the event listeners
            var bExecuteDefault = oTable.fireFilter({
                column: this,
                value: sValue
            });

            if (bExecuteDefault) {

                this.setProperty("filtered", !!sValue, true);
                this.setProperty("filterValue", sValue, true);
                var oMenu = this.getMenu();
                var ColumnMenu = sap.ui.require("sas/hc/ui/table/ColumnMenu");
                if (oMenu && ColumnMenu && oMenu instanceof ColumnMenu) {
                    // update column menu input field
                    oMenu._setFilterValue(sValue);
                }

                var aFilters = [];
                var aCols = oTable.getColumns();
                for (var i = 0, l = aCols.length; i < l; i++) {
                    var oCol = aCols[i],
                        oFilter;

                    oMenu = oCol.getMenu();
                    try {
                        oFilter = oCol._getFilter();
                        if (oMenu && ColumnMenu && oMenu instanceof ColumnMenu) {
                            oMenu._setFilterState(ValueState.None);
                        }
                    } catch (e) {
                        if (oMenu && ColumnMenu && oMenu instanceof ColumnMenu) {
                            oMenu._setFilterState(ValueState.Error);
                        }
                        continue;
                    }
                    if (oFilter) {
                        aFilters.push(oFilter);
                    }
                }
                oTable.getBinding("rows").filter(aFilters, FilterType.Control);

                this._updateIcons();

            }

        }

        return this;

    };

    /**
     * @private
     */
    Column.prototype._parseFilterValue = function(sValue) {
        var oFilterType = this.getFilterType();

        if (oFilterType) {
            if (jQuery.isFunction(oFilterType)) {
                sValue = oFilterType(sValue);
            } else {
                sValue = oFilterType.parseValue(sValue, "string");
            }
        }

        return sValue;
    };

    /**
     * @private
     */
    Column.prototype._restoreIcons = function() {
        this._updateIcons();
    };

    /**
     * Returns whether the column should be rendered or not.
     * @return {boolean} true, if the column should be rendered
     * @protected
     */
    Column.prototype.shouldRender = function() {
        return this.getVisible() && !this.getGrouped();
    };

    Column.PROPERTIES_FOR_ROW_INVALIDATION = {visible: true, required: true, flexible: true, headerSpan: true};
    /*
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setProperty = function(sName, vValue) {
        var oTable = this.getParent();

        // if we're setting the width property and it's the same as
        // the old width, return now so we don't invalidate unnecessarily
        if (sName === "width" && this.getProperty(sName) === vValue) {
            return;
        }

        if (oTable &&
            oTable._resetRowTemplate &&
            this.getProperty(sName) !== vValue &&
            Column.PROPERTIES_FOR_ROW_INVALIDATION[sName] &&
            (this.getVisible() || sName === "visible" || sName === "required")) {

            oTable._resetRowTemplate();
        }

        return Element.prototype.setProperty.apply(this, arguments);
    };

    /*
     * support the declarative usage of the filter type
     * @see JSDoc generated by SAPUI5 control API generator
     */
    Column.prototype.setFilterType = function(vType) {
        var oType = vType;
        if (typeof (vType) === "string") {
            try {
                // similar to BindingParser allow to specify formatOptions and constraints for types
                var mConfig = jQuery.sap.parseJS(vType);
                if (typeof (mConfig.type) === "string") {
                    var fnType = jQuery.sap.getObject(mConfig.type);
                    oType = fnType && new fnType(mConfig.formatOptions, mConfig.constraints);
                }
            } catch (ex) {
                var fnType = jQuery.sap.getObject(vType);
                oType = fnType && new fnType();
            }
            // check for a valid type
            if (!(oType instanceof Type)) {
                jQuery.sap.log.error("The filter type is not an instance of sap.ui.model.Type! Ignoring the filter type!");
                oType = undefined;
            }
        }
        this.setProperty("filterType", oType, true);
        return this;
    };

    /**
     * Determines the column index based upon the order in its aggregation.
     * Invisible columns are taken in account of order.
     * @see JSDoc generated by SAPUI5 control API generator
     * @return {int} the column index.
     * @public
     */
    Column.prototype.getIndex = function() {
        var oTable = this.getParent();
        if (oTable) {
            return oTable.indexOfColumn(this);
        } else {
            return -1;
        }
    };

    /**
     * Create a Sorter object for the given column and return it by way of a Promise
     * @protected
     * @param {object} oCol Associated column object
     * @returns {object} a Promise that will resolve with the new Sorter object
     */
    Column.prototype._createSorterPromise = function(oColumn) {
        var oCol = oColumn,
            bCaseSensitiveSorterArg, oSorter;

        // default to this column
        if (!oColumn) {
            oCol = this;
        }

        bCaseSensitiveSorterArg = oCol.getCaseSensitiveSort();

        // if consumer has not specified a custom sort comparator, use ltjs collator
        if (typeof oCol.getSortFunction() !== "function") {
            bCaseSensitiveSorterArg = false; // collator will handle this, rather than SorterProcessor
            var sStrength = "secondary";
            if (oCol.getCaseSensitiveSort() === true) {
                sStrength = "tertiary";
            }

            // LTJS returns a Promise of a sort comparator, so construct our own Promise that resolves with our Sorter
            var oPromise = new Promise(function(fnResolve, fnReject) {
                // use SASCollatorUtil to get the sort comparator
                sas.ltjs.sasicu.util.SASCollatorUtil.getComparatorPromise(null, sStrength).then(function(fnComparator) {
                    // if dealing with strings use collator, otherwise use default comparator
                    oCol.setProperty("sortFunction", function(s0, s1) {
                        //if either argument is a string, then use the collator's comparator
                        if (typeof s0 === "string" || typeof s1 === "string") {
                            return fnComparator(s0, s1);
                        }

                        //for all other data types, use the default comparator
                        return sap.ui.model.Sorter.defaultComparator(s0, s1);
                    }, true);   //avoid invalidation

                    // create the new Sorter based on our sort comparator and other sort properties from column
                    oSorter = new sap.ui.model.Sorter({
                        path: oCol.getSortProperty(),
                        descending: oCol.getSortOrder() === sas.hc.ui.table.SortOrder.Descending,
                        caseSensitive: bCaseSensitiveSorterArg,
                        fnCompare: oCol.getSortFunction()
                    });

                    // resolve the promise with the new Sorter
                    fnResolve(oSorter);
                });
            });
            return oPromise;
        }

        // if we already have a sort function, then simply return a resolved Promise with the new Sorter
        oSorter = new sap.ui.model.Sorter({
            path: oCol.getSortProperty(),
            descending: oCol.getSortOrder() === sas.hc.ui.table.SortOrder.Descending,
            caseSensitive: bCaseSensitiveSorterArg,
            fnCompare: oCol.getSortFunction()
        });

        return Promise.resolve(oSorter);
    };

    /**
     * Setter for the "highlighted" property
     * @public
     */
    Column.prototype.setHighlighted = function( bHighlighted ) {
        this.setProperty('highlighted', bHighlighted, true);
        return this;
    };

    /**
     * Returns the array of sap.ui.model.Sorter objects in the table's "rows" binding or null if there is no binding.
     * @returns {*}
     * @private
     */
    Column.prototype._getBoundSorters = function() {
        var oTable = this.getParent(),
            oBinding,
            aSorters = null;

        if (!!oTable) { //possible to not have a parent table
            oBinding = oTable.getBinding("rows");
            if (!!oBinding) {   //possible to not be bound
                aSorters = oBinding.aSorters;
            }
        }

        return aSorters;
    };

    /**
     * Returns true if this column is both sorted and represents the primary parameter involved in sorting the data.
     * @returns {boolean} true if primary sort, false otherwise
     * @public
     */
    Column.prototype.isPrimarySort = function() {
        var aSorters = this._getBoundSorters(),
            bPrimarySort = false;

        if (aSorters !== null && aSorters.length > 0 && aSorters[0].sPath === this.getSortProperty()){
            bPrimarySort = true;
        } else if (this.getSorted() === true && aSorters !== null && aSorters.length === 0) {
            //we're probably sorting the first column right now
            bPrimarySort = true;
        }

        return bPrimarySort;
    };

    /**
     * Returns a string used in javascript to determine the nature of the sort icon being displayed.  There are 2
     * key factors to include - ascending sort order vs descending and primary sort parameter vs secondary
     * @param bAscending true if the string should represent ascending sort order, false for descending
     * @param bPrimary true if the string should represent primary sort parameter, false for secondary
     * @returns {string} the string that can be used to tell what kind of sort is being indicated
     * @private
     */
    Column.prototype._getSortTypeMapString = function(bAscending, bPrimary) {
        if (bAscending === undefined) {
            bAscending = (this.getSortOrder() === sas.hc.ui.table.SortOrder.Ascending);
        }
        if (bPrimary === undefined) {
            bPrimary = this.isPrimarySort();
        }
        return (bAscending ? "A" : "D") + (bPrimary ? "1" : "2");
    };

    /**
     * hook function for _renderSortIcon - renders the sort icon into the icon container if the given icon is not the one
     * desired. if the given sort icon is not the eventual icon, then the existing one will be animated as it changes
     * @protected
     */
    Column.prototype._renderCustomSortIcon = function($sortIcon) {
        //helper function to animate the rotation of the icon
        function fnRotate($el, degree, fnAfter, iDuration) {
            $({deg: 0}).animate({
                deg: degree
            }, {
                step: function(now,fx) {
                    $el.css('-webkit-transform','rotate('+now+'deg)');
                    $el.css('-moz-transform','rotate('+now+'deg)');
                    $el.css('transform','rotate('+now+'deg)');
                },
                duration: (iDuration || AnimationUtil.DURATION_FAST),
                always: fnAfter
            });
        }

        // create the image for the sort order visualization
        // S1121468 - using SAS wrapped image instead of raw SAP icon control
        var bPrimarySort = this.isPrimarySort();

        var self = this,
            oSortOrder = this.getSortOrder(),
            oNewSrc = ((oSortOrder === sas.hc.ui.table.SortOrder.Ascending)
                ? (bPrimarySort ? sas.icons.HC.SORTARROWPRIMARYUP : sas.icons.HC.SORTARROWSECONDARYUP)
                : (bPrimarySort ? sas.icons.HC.SORTARROWPRIMARYDOWN : sas.icons.HC.SORTARROWSECONDARYDOWN)
            ),
            sSortType = this._getSortTypeMapString((oSortOrder === sas.hc.ui.table.SortOrder.Ascending),(bPrimarySort)),
            fnAfterAnimation = function() {
                var domIcons = self.getDomRef("icons");
                if (!domIcons) {
                    return; //no DOM for icons, so abort - invalidation or table destruction has interrupted us
                }
                // apply the image and aria property to the column
                var oRenderManager = sap.ui.getCore().createRenderManager(),
                    $self = self.$();

                oRenderManager.writeIcon(oNewSrc, ["sasSortIcon"], { "data-sas-sort": sSortType });   //writeIcon(oSrc, aClasses, mAttributes);
                $self.find(".sasSortIcon").remove(); //remove any existing sort icon
                // S1121465 - sort icon should be after filter icon
                oRenderManager.flush(domIcons, true, (self.getFiltered()));  //flush(oTargetDomNode, bDoNotPreserve, vInsert)
                $self.attr("aria-sort", oSortOrder === sas.hc.ui.table.SortOrder.Ascending ? "ascending" : "descending");
            };

        if (!!$sortIcon && $sortIcon.length > 0 && $sortIcon.attr("data-sas-sort") !== sSortType && AnimationUtil.getVisualEffectsEnabled()) {  //check to see if the icon exists and needs animation
            fnRotate($sortIcon, (oSortOrder === sas.hc.ui.table.SortOrder.Ascending) ? -180 : 180, fnAfterAnimation);
        } else {
            fnAfterAnimation(); //just do the afterward part w/o animation
        }
    };

    /**
     * Render filter icon
     * Rather than using an Image type like SAP, we use an Icon type
     * @override
     * @private
     */
    Column.prototype._renderFilterIcon = function() {
        var oTable = this.getParent();
        if (oTable && oTable.getDomRef()) {     //the table needs to be in the DOM
            var domIcons = this.getDomRef("icons");
            if (domIcons !== null) {    //along with the icon container
                var oRenderManager = sap.ui.getCore().createRenderManager();
                oRenderManager.writeIcon(sas.icons.HC.FILTER, ["sasFilterIcon"] /*, mAttributes*/);
                oRenderManager.flush(domIcons);
            }
        }
    };

    /**
     * Setter for the "headerEditable" property
     * @param bEditable whether the header should be editable
     * @public
     */
    Column.prototype.setHeaderEditable = function(bEditable) {
        var sValue = this.getHeaderText(),
            oLabel;

        if (bEditable === this.getHeaderEditable()) {
            return;
        }

        if(sValue !== undefined) {
            if(bEditable) {
                oLabel = this.getLabel();   //fetch any existing label control
                this.setIpeValue(sValue);
                //if we have a label that is not already the IPE, see if it has a 'textAlign' property
                if (oLabel && oLabel !== this._oInPlaceEdit && jQuery.isFunction(oLabel.getTextAlign)) {
                    //apply that textAlign property to the IPE before replacing the label and losing that info
                    this._oInPlaceEdit.setTextAlign(oLabel.getTextAlign());
                }
                this.setLabel(this._oInPlaceEdit);
            }else {
                this.setLabel(sValue);
            }
        }


        return this.setProperty("headerEditable", bEditable);
    };

    //added for HTMLCOMMONS-4582 - convenience method for getting some sort of text label for the column header
    //  * If we can remove this and keep it entirely with the Table, that would be better for backward compatibility with the SAP Column class
    // TODO There is no longer an SAP Column that we use. Should revist above comment.
    Column.prototype.getHeaderText = function() {
        var oTable = this.getParent();
        if (oTable) {
            return oTable.getColumnHeaderText(this);
        }

        var sValue, oControl = this.getLabel();
        if(!!oControl) {
            if(oControl.getText) {
                sValue = oControl.getText();
            }else if(oControl === this._oInPlaceEdit) {
                sValue = oControl.getContent().getValue();
            }else if(oControl.getTooltip) {
                sValue = oControl.getTooltip();
            }
        }

        return sValue;
    };

    Column.prototype.setIpeValue = function(sValue) {
        var oContent = this._oInPlaceEdit.getContent();
        oContent.setValue(sValue);
        return this;
    };

    //auto-generate a label control if a string was passed (as a convenience)
    Column.prototype.setSummaryLabel = function(vLabel) {
        var oLabel = vLabel;
        if (typeof (vLabel) === "string") {
            oLabel = sas.hc.ui.table.TableHelper.createLabel({text: vLabel});
        }
        this.setAggregation("summaryLabel", oLabel);
        return this;
    };

    /**
     * Returns true if this Column's width property is
     * considered an absolutely sized width.
     *
     * @see {@link TableUtils.isAbsoluteWidth} for more details.
     *
     * @return True if this Column has an absolute width, false otherwise.
     */
    Column.prototype.hasAbsoluteWidth = function() {
        return  TableUtils.isAbsoluteWidth(this.getWidth());
    };

    Column.prototype._restoreOriginalWidth = function() {
        this.setWidth(this._sConfiguredWidth);
    };

    /**
     * Determine if an edge of the column is past the visible part of the scroll container.
     * Takes summary headers into account if column header is hidden.
     * Handles RTL internally.
     * This function is meant to be consumed by the _is(Near|Far)EdgeWithinScrollRange functions
     * @param {boolean} bFar whether to check the far or near column
     * @return {boolean} if the far-edge of the column is visible
     * @private
     */
    Column.prototype._isEdgeWithinScrollRange = function(bFar) {
        var oTable = this.getParent(),
            $summaryHeader, iColumnLeftOffset, iColumnOuterWidth;

        if (oTable.isFrozenColumn(this) === true) {
            return false;
        }

        if (this.$().length === 0) {
            return false;
        }

        if (oTable.getShowSummaryHeader() === true) {
            $summaryHeader = oTable.$().find('.sasUiTableColSumHead').eq(this.getIndex());
            iColumnLeftOffset = $summaryHeader.offset().left;
            iColumnOuterWidth = $summaryHeader.outerWidth();
        } else {
            iColumnLeftOffset = this.$().offset().left;
            iColumnOuterWidth = this.$().outerWidth();
        }

        // opposite for RTL
        if (oTable._bRtlMode === true) {
            bFar = !bFar;
        }

        if (bFar === true) {
            return oTable._isXCoordWithinScrollRange(iColumnLeftOffset + iColumnOuterWidth);
        }
        return oTable._isXCoordWithinScrollRange(iColumnLeftOffset);
    };

    /**
     * Determine if the far edge of the column (right-edge in LTR) is past the visible part of the scroll container.
     * Takes summary headers into account if column header is hidden
     * @return {boolean} if the far-edge of the column is visible
     * @private
     */
    Column.prototype._isFarEdgeWithinScrollRange = function() {
        return this._isEdgeWithinScrollRange(true);
    };

    /**
     * Determine if the near edge of the column (left-edge in LTR) is past the visible part of the scroll container.
     * Takes summary headers into account if column header is hidden
     * @return {boolean} if the near-edge of the column is visible
     * @private
     */
    Column.prototype._isNearEdgeWithinScrollRange = function() {
        return this._isEdgeWithinScrollRange(false);
    };

    /**
     * Determine if both the near and far edges of the column are outside of the visible part of the scroll container
     * @private
     * @return {boolean} if both edges of the column are not visible
     */
    Column.prototype._areBothEdgesOutsideOfScrollRange = function() {
        return this._isNearEdgeWithinScrollRange() === false && this._isFarEdgeWithinScrollRange() === false;
    };

    /**
     * Listen to keyboard events to perform a column resize operation
     *
     * Wrap the resize operation in a promise so it can be clear when the operation starts and stops.
     * Only one resize operation is allowed at a time.
     *
     * When the resize operation starts:
     * * listen to keydown -> cursor keys for resizing, esc to cancel, enter/tab to finish
     * * listen to Column's blur -> to stop resize operation
     * * listen to horizontal scrolling -> to stop resize operation if column goes off-screen
     * * prevent typical item nav keyboard handlers
     *
     * The resize operation will complete when:
     * * User hits esc/enter/tab
     * * column being resized recieves blur event
     * * Horizontal scrolling occurs leaving the column no longer visible
     *
     * When the resize operation completes (promise resolves):
     * * resume typical item nav keyboard handlers
     * * clean up event handlers
     * * sync Table headers/rows
     *
     * @return {object} promise to be resolved when the keyboard resize operation is complete
     * @private
     */
    Column.prototype._resizeWithKeyboard = function() {
        var self = this,
            oTable = this.getParent(),
            // the reference representing the column
            $colRef = this._getKeyboardResizeReference(),
            // used when a resize operation is cancelled
            iOriginalWidth = $colRef.width(),
            // used to prevent resize operations from starting during column width bumping (which could lead to column nudging)
            bDuringColumnWidthBump = false,
            // keep track if any resizing actually occurred or not
            bDidAnyResizingOccur = false,
            // used as a hookpoint for when Table invalidates
            oTableDelegate;

        oTable._bDuringResizeWithKeyboard = true;

        // prevent normal keyboard navigation during resize operation
        oTable._getKeyboardExtension()._itemNavigationSuspended = true;

        return new Promise(function(resolve) {
            // listen to keydown and blur events
            // resolve when column resizing operation is complete
            (new Promise(function(resolve) {
                self._startKeyboardResizing(true);
                jQuery(window).on('keydown.tbl_keyboard_resize', function(oEvent) {
                    oEvent.stopPropagation();
                    oEvent.preventDefault();
                    if (bDuringColumnWidthBump !== false) {
                        return;
                    }

                    switch (oEvent.keyCode) {
                        case jQuery.sap.KeyCodes.ARROW_RIGHT:
                            if (oTable._bRtlMode === true) {
                                oTable._bumpDownColumnWidth(self);
                            } else {
                                bDuringColumnWidthBump = true;
                                oTable._bumpUpColumnWidth(self).then(function(e) {
                                    bDuringColumnWidthBump = false;

                                    // certain conditions regarding fixed columns can
                                    // cause the resize operation to be stopped. See S1373041
                                    if (oTable._bIgnoreFixedColumnCount === true) {
                                        resolve();
                                    }
                                });
                            }
                            bDidAnyResizingOccur = true;
                            break;

                        case jQuery.sap.KeyCodes.ARROW_LEFT:
                            if (oTable._bRtlMode === true) {
                                bDuringColumnWidthBump = true;
                                oTable._bumpUpColumnWidth(self).then(function() {
                                    bDuringColumnWidthBump = false;

                                    // certain conditions regarding fixed columns can
                                    // cause the resize operation to be stopped. See S1373041
                                    if (oTable._bIgnoreFixedColumnCount === true) {
                                        resolve();
                                    }
                                });
                            } else {
                                oTable._bumpDownColumnWidth(self);
                            }
                            bDidAnyResizingOccur = true;
                            break;

                        case jQuery.sap.KeyCodes.ESCAPE:
                            oTable._updateColumnWidth(self, iOriginalWidth);
                        case jQuery.sap.KeyCodes.TAB:
                        case jQuery.sap.KeyCodes.ENTER:
                            resolve();
                            break;

                        default:
                            break;
                    }
                });

                // if the scroll-bars are clicked upon stop operation
                jQuery(oTable.getDomRef(SharedDomRef.HorizontalScrollBar)).on("mousedown.tbl_keyboard_resize", resolve);
                jQuery(oTable.getDomRef(SharedDomRef.VerticalScrollBar)).on("mousedown.tbl_keyboard_resize", resolve);

                // Listen to when focus leaves the column dom ref (blur).
                // When Table invalidates the blur dom ref reference is lost, so re-bind the blur event when necessary
                $colRef.on("blur.tbl_keyboard_resize", resolve);
                oTableDelegate = {
                    onAfterRendering: function() {
                        // re-set the blur listener
                        $colRef = self._getKeyboardResizeReference();
                        $colRef.off("blur.tbl_keyboard_resize"); // probably a no-op, since reference is already gone
                        $colRef.on("blur.tbl_keyboard_resize", resolve);

                        if (self._areBothEdgesOutsideOfScrollRange() === true) {
                            resolve();
                        }

                        // re-set scroll-bar click handlers
                        jQuery(oTable.getDomRef(SharedDomRef.HorizontalScrollBar)).off("mousedown.tbl_keyboard_resize");
                        jQuery(oTable.getDomRef(SharedDomRef.VerticalScrollBar)).off("mousedown.tbl_keyboard_resize");
                        jQuery(oTable.getDomRef(SharedDomRef.HorizontalScrollBar)).on("mousedown.tbl_keyboard_resize", resolve);
                        jQuery(oTable.getDomRef(SharedDomRef.VerticalScrollBar)).on("mousedown.tbl_keyboard_resize", resolve);
                    }
                };
                oTable.addEventDelegate(oTableDelegate);
            })).then(function() {
                jQuery(oTable.getDomRef(SharedDomRef.HorizontalScrollBar)).off("mousedown.tbl_keyboard_resize");
                jQuery(oTable.getDomRef(SharedDomRef.VerticalScrollBar)).off("mousedown.tbl_keyboard_resize");
                jQuery(window).off("keydown.tbl_keyboard_resize");
                $colRef.off("blur.tbl_keyboard_resize");

                oTable.removeEventDelegate(oTableDelegate);
                self._startKeyboardResizing(false);
                oTable._getKeyboardExtension()._itemNavigationSuspended = false;

                // only call syncing if a resize actually occurred
                if (bDidAnyResizingOccur === true) {
                    oTable._syncColumnHeaders(oTable._collectTableSizes());
                    oTable._resizeDependentColumns(self, self.getWidth());
                    oTable.invalidate();
                }

                self._bDuringResizeWithKeyboard = false;
                resolve();
            });
        });
    };

    Column.prototype.getRequired = function() {
        return this.getMandatory() || this.getProperty("required");
    };

    Column.prototype.setMandatory = function(bMandatory, bSuppressInvalidate) {
        this.setProperty("mandatory", bMandatory, bSuppressInvalidate);

        var oParent = this.getParent();
        if (oParent && jQuery.isFunction(oParent._enforceMandatoryColumnOrder)) {
            oParent._enforceMandatoryColumnOrder();
        }
    };

    return Column;
});
