/*****************************************************************/
/* Copyright (c) 2019 by SAS Institute Inc., Cary, NC, USA.      */
/* All Rights Reserved.                                          */
/*****************************************************************/
/* NAME:     rskmgt.sas                                          */
/* DESCRIPTION: Initializes macro paths for SAS Risk Governance  */
/*              Framework                                        */
/* PARAMETER: RGF_CONN_LASR = connect to the RGF LASR Server     */
/*            YES or NO  Default: YES                            */
/* PRODUCT:  SAS Risk Governance Framework                       */
/*****************************************************************/

%macro rskmgt(CONN_LASR=YES);

  %put NOTE: {START: &sysmacroname};

  /* declare RGF globals */
  %global RGF_SETUP;
  %global RGF_STATUS;
  %global RGF_CONN_LASR;

  /* assign the CONN_LASR value to a global macro */
  %let RGF_CONN_LASR=&CONN_LASR;

  /* mark status as unknown */
  %let RGF_STATUS = 1;

  /* only perform setup once per sas session */
  %if &RGF_SETUP ^= %str() %then
  %do;
    %let RGF_STATUS = 0;
    %goto DONE;
  %end;

  %local RGF_MACRO_PATHS ORIGINAL_SASAUTOS CONTENT_UCMACRO_PATHS;
  %global RGF_CUSTOM_PATH;
  %global RGF_SWC_MID_CLASSID RGF_SWC_SVRC_CLASSID;

  %let RGF_SWC_MID_CLASSID=rgpmid;
  %let RGF_SWC_SVRC_CLASSID=rgpsvrc;

  /***************************************/
  /* Get the Root DAV path from metadata */
  /***************************************/

  /* Get the content server URL */
  %if not %symexist(service_endpoint) %then %global service_endpoint;
  %svcreg(type=REST, name=ServerService REST);
  %let dav_srv=&service_endpoint;

  /* Parse the URL to keep just the server and port information */
  %global srv_url;
  %let srv_url=%substr(&dav_srv,1,(%sysfunc(indexw("&dav_srv",%qscan(%bquote(&dav_srv),3,'/'),"/"))-3));

  /* Get the remainder of the DAV path from SMC */
  %rskmgt_get_app_property_value(application=&RGF_SWC_MID_CLASSID,
                                 property=Dav.url.path,
                                 attribute=ClassIdentifier,
                                 output_property_value=dav_path,
                                 output_status=dav_status);

  %let dav_path=%trim(%nrstr(&dav_path));

  /* Put the content server URL and DAV path together */
  %global dav_url;
  %let dav_url=&srv_url.&dav_path;

  /***********************************************************************************/
  /* Read the content-registry.properties file to see what content ids are installed */
  /* and get the schema names for each content ids.                                  */
  /***********************************************************************************/

  filename svc_rsp temp;

  proc http url="&dav_url./rgf-system/Config/content-registry.properties"
            out=svc_rsp
            method="get"
            ct="application/x-www-form-urlencoded"
            HTTP_TOKENAUTH;
  run;

  %let termstr=;

  data _null_;
    infile svc_rsp lrecl=32767;
    input @;
    if index(_infile_,'0D'x)>0 then call symput('termstr','CR');
    stop;
  run;

  data cntreg_props;
    infile svc_rsp recfm=v
    %if %symexist(termstr) %then
    %do;
      %if &termstr ne %str() %then
      %do;
        termstr=&termstr;
      %end;
      %else
      %do;
        %str(;)
      %end;
    %end;
    %else
    %do;
      %str(;)
    %end;

    input x $char2000. @@;
  run;

  filename svc_rsp;

  %local nobs rc;
  %let dsid = %sysfunc(open(work.cntreg_props));
  %if &dsid %then
  %do;
    %let nobs  = %sysfunc(attrn(&dsid, nlobs));
    %let rc = %sysfunc(close(&dsid));
  %end;
  %else %let nobs  = 0;

  %if &nobs > 0 %then
  %do;
    data cntreg_props;
      set cntreg_props;
      x=unicode(x,'esc');
    run;

    /* Creates a macro variable to hold the content ids */
    data _null_ other_keys;
      set cntreg_props(where=(x contains "="));
      length key text content_ids $100;
      %if &termstr ne %str() %then
      %do;
        x=compress(x,,"c");
      %end;
      key=scan(upcase(x),1,"=");
      text=scan(x,-1,"=");

     if substr(key,1,1)='#' then delete;

      if findw(key,'RGF.CONTENT.IDS')>0 and substr(key,1,1) ne '#' then
      do;
        content_ids=text;
        call symputx('rgf_content_ids',text,'G');
      end;
      else output other_keys;
    run;

     /* Loop through the other_keys dataset and get the content id schema names */
     %let z=1;
     %let contentId=%trim(%qscan(%bquote(%upcase(&rgf_content_ids)),&z,%str(,)));
     %do %while (&contentId ne );

       data _null_;
         set other_keys;
         if findw(key,"RGF.CONTENT.&contentId..SCHEMA")>0 then
         do;
           if text = '' then
           do;
             call symputx("&contentId._schema",&contentID,'G');
           end;
           if upcase(text) = '<<DEFAULT>>' then
           do;
             call symputx("&contentId._schema",'','G');
           end;
           if upcase(text) ne '<<DEFAULT>>' and text ne '' then
           do;
             call symputx("&contentId._schema",text,'G');
           end;
         end;
       run;

       %global &contentId._schema;
       %if &&contentId._schema eq %str() %then %let &contentId._schema=;

       %let z=%eval(&z+1);
       %let contentId=%trim(%qscan(%bquote(%upcase(&rgf_content_ids)),&z,%str(,)));
     %end;
   %end;

  /* add our UCMACRO paths to SASAUTOS */
  /* Get the RGF Custom Ucmacros directory path from RGF_SWC_SVRC_CLASSID */
  %rskmgt_get_app_property_value(application=&RGF_SWC_SVRC_CLASSID,
                                 property=rgf.custom.macro.path,
                                 attribute=ClassIdentifier,
                                 output_property_value=RGF_CUSTOM_PATH,
                                 output_status=rgf_status);

  %let RGF_CUSTOM_PATH=%trim(&RGF_CUSTOM_PATH);

  /* Lookup the RGF ucmacros path in metadata and build the ucmacros path for the content ids installed */
  %local CONTENT_UCMACRO_PATHS;
  %let CONTENT_UCMACRO_PATHS=;

  %let z=1;
  %let contentId=%trim(%qscan(%bquote(&rgf_content_ids),&z,%str(,)));
  %do %while (&contentId ne );
    %let &contentId.UcmacrosPath=%substr(&RGF_CUSTOM_PATH,1,(%sysfunc(indexw("&RGF_CUSTOM_PATH",%qscan(%bquote(&RGF_CUSTOM_PATH),-2,'/\'),"/\"))-2))Source/&contentId./ucmacros;

    %if &z=1 %then
    %do;
      %let CONTENT_UCMACRO_PATHS="&&&contentId.UcmacrosPath";
    %end;
    %else
    %do;
      %let CONTENT_UCMACRO_PATHS=&CONTENT_UCMACRO_PATHS "&&&contentId.UcmacrosPath";
    %end;

    %let z=%eval(&z+1);
    %let contentId=%trim(%qscan(%bquote(&rgf_content_ids),&z,%str(,)));
  %end;

  /* prepend the macro paths to SASAUTOS, removing any next parentheses and commas */
  %let ORIGINAL_SASAUTOS= %sysfunc(compress(%sysfunc(getoption(sasautos)),%str(%(%))));

  %if ((&SYSSCP eq WIN) or (&SYSSCP eq DNTHOST)) %then
  %do;
    %let rgf_macro_paths = !SASROOT/rskmgtsvr/ucmacros;
  %end;
  %else
  %do;
    %let rgf_macro_paths = !SASROOT/ucmacros/rskmgtsvr;
  %end;

  %local new_autos;
  %let new_autos=&CONTENT_UCMACRO_PATHS "&RGF_CUSTOM_PATH" "&RGF_MACRO_PATHS" &ORIGINAL_SASAUTOS;
  options sasautos=(&new_autos) dbidirectexec compress=yes;
  %put rskmgt set SASAUTOS to: &new_autos;

  /* Include any site specific post init setup */
  %rgf_post_init;

  /* Mark status as good */
  %let RGF_STATUS = 0;

  %DONE:

  /* Mark setup complete */
  %let RGF_SETUP = 1;

  %put NOTE: {END: &sysmacroname};

%mend rskmgt;

/* Macro used to lookup properties for a software component */
%macro rskmgt_get_app_property_value(application=,
                                     attribute=Name,
                                     property=,
                                     output_property_value=,
                                     output_status=);

%put NOTE: {START: &sysmacroname};

 %global &output_property_value &output_status;

 data _null_;
   error = 0;

   length property_value $256;
   property_value = '';

   length uri_software_component $256;
   uri_software_component='';
   software_component_found = metadata_getnobj("omsobj:SoftwareComponent?@&attribute='&application'",
                                               1, uri_software_component);

   if software_component_found = 1 then do;

     application_configuration_found = 0;    /*Application Configuration PropertySet found*/

     property_sets_rc = 1;
     property_sets_count = 1;
     length uri_application_configuration $256;
     uri_application_configuration = '';
     do while (property_sets_rc > 0);
       property_sets_rc = metadata_getnasn(uri_software_component, "PropertySets",
                                            property_sets_count, uri_application_configuration);
       if property_sets_rc > 0 then do;
         length property_set_name $256;
         property_set_name = '';
         rc2 = metadata_getattr(uri_application_configuration, "Name", property_set_name);
         if property_set_name = 'Application Configuration' then do;
           application_configuration_found=1;
           property_sets_rc = 0;  /*don't need to keep searching through PropertySets*/
           set_properties_rc = 1;
           set_properties_count = 1;
           length uri_property $256;
           uri_property = '';
           property_found = 0;
           do while (set_properties_rc > 0);
             set_properties_rc = metadata_getnasn(uri_application_configuration, "SetProperties",
                                            set_properties_count, uri_property);
             if set_properties_rc > 0 then do;
               length property_name $256;
               property_name = '';
               rc2 = metadata_getattr(uri_property, "Name", property_name);
               /*%if &debug_rolcap = Y %then %do;
                 put property_name;
               %end;*/
               if property_name = "&property" then do;
                 property_found = 1;
                 *put property_name;
                 set_properties_rc = 0;  /*don't need to keep searching through SetProperties*/
                 rc2 = metadata_getattr(uri_property, "DefaultValue", property_value);
                 *put property_value;
               end;
             end;
             else if set_properties_rc eq -4 and set_properties_count eq 1 then do;
               put 'SetProperties Association not found.';
               error = 4;
               leave;
             end;
             set_properties_count = set_properties_count + 1;
           end;   /* going through SetProperties*/

           if property_found eq 0 then do;
             put 'Property not found.';
             if error eq 0 then error = 5;
           end;

         end;   /* if Application Configuration PropertySet found */

       end;
       else if property_sets_rc eq -4 and property_sets_count eq 1 then do;
         put 'PropertySets Association not found.';
         error = 2;
         leave;
       end;
       property_sets_count = property_sets_count + 1;
     end;   /* going through the PropertySets */

     if application_configuration_found eq 0 then do;
       put 'Application Configuration PropertySet not found.';
       if error eq 0 then error = 3;
     end;

   end;
   else do;
     put "&application SoftwareComponent not found.";
     put 'No metadata extracted.';
     error = 1;
   end;

   call symput("&output_status",left(put(error,3.)));
   call symput("&output_property_value",property_value);

  run;

%put NOTE: {END: &sysmacroname};

%mend rskmgt_get_app_property_value;
