/**********************************************************************/
/*  Copyright (c) 2019 by SAS Institute, Inc., Cary, NC, USA.         */
/*  All rights reserved.                                              */
/**********************************************************************/
/*  NAME:        rskmgt_apply_dim_path.sas                            */
/*  VERSION:     7.4                                                  */
/*  DESCRIPTION: Apply dimension paths to the input table.            */
/*  PARAMETERS:  indsn=target table                                   */
/*               feature=target feature                               */
/*               prefix=prefix used in the dimension names when mult. */
/*                     dimension locations are needed in the output   */
/*                     (ex. incidents may want the discovery and event*/
/*                          location in the output table)             */
/*               prefix_label=Displayable name for the prefix parm    */
/*                            (ex. incidents would be "Discovery" for */
/*                                 "Discovery" location)              */
/*               keep_dimension=dimension to keep in the table.       */
/*                                   (used by CAUSE and CONTROL to    */
/*                                    keep only those dimensions for  */
/*                                    those objects)                  */
/*  PRODUCT:     SAS Risk Management Server                           */
/**********************************************************************/

%macro rskmgt_apply_dim_path(indsn=, 
                             feature=,
                             prefix=,
                             prefix_label=,
                             keep_dimension=);

%put NOTE: {START: &sysmacroname};
     
/* Generate a filtered list of dimensions and levels based on the &indsn */
%rskmgt_ds_exists(indsn=work.dims2,
                  output_nobs=dims_cnt);

%if &dims_cnt = 0 %then %goto exit;

/* Create a table of all dimensions w/ dimensional point values */
data work.filtered_dims;
  length dim_level 8. dimension_nm $400;
  set work.dims2;
       
  if scan(vname,-1,'_RK')=. then dim_level=.;
  else dim_level=scan(vname,-1,'_');
  dimension_nm=substr(vname,1,int(index(vname,'_RK')) - 1);
  
  if vname='DIM_POINT_RK' then delete;
run;

/* Add a dim_level=1 for the dimensions with just root node values */
proc sort data=work.filtered_dims out=work.filtered_dims_srt;
  by dimension_nm;
run;

data work.filtered_dims;
  set work.filtered_dims_srt;
  by dimension_nm;
  
  if first.dimension_nm and last.dimension_nm then dim_level=1;
run;

/* Sort by the vname to get the dimension values in the correct order */
proc sort tagsort data=filtered_dims;
  by dimension_nm dim_level;
run;
 
/* Filter feature_dimensions by feature and merge w/ filtered dimensions */
proc sql;
  create table work.feature_dims_lst as
  select *
  from &contentId.ctl.feature_dimensions
  where upcase(feature_nm)=%upcase("&feature")
      
  %if &keep_dimension ne %str() %then
  %do;
    and upcase(dimension_nm) = "%upcase(&keep_dimension)"
  %end;
  ;
quit;

/* Join the filtered dimensions and feature dimensions to get additional attributes */
proc sql;
  create table work.feature_filtered_dims as
  select a.*, b.dimension_alias_nm
  from work.filtered_dims as a left join
       work.feature_dims_lst as b
  on a.dimension_nm=b.dimension_nm
  order by a.dimension_nm, a.dim_level;
quit;

/* Create the DIMCOUNT, DIMENSION1-DIMENSIONX and DIMALIAS1-DIMALIASX macro variables */
proc sql noprint;
  select count(distinct dimension_nm) into : DIMCOUNT trimmed
  from work.feature_filtered_dims;
  
  select distinct dimension_nm, dimension_alias_nm  into : DIMENSION1 - :DIMENSION&DIMCOUNT,
                                                         : DIM_ALIAS1 - :DIM_ALIAS&DIMCOUNT
  from work.feature_filtered_dims
  where dim_level ^=.
  order by dimension_nm;
quit;

  /* Begin dimension format generation */
  %macro rskmgt_create_dim_path_format(dim_nm=&&DIMENSION&I);
    
    %put NOTE: {START: &sysmacroname};
  
    %global &dim_nm._lvls;

    proc sql noprint;
       select max(dim_level) into : &dim_nm._lvls trimmed
       from work.feature_filtered_dims
       where dimension_nm="&dim_nm";
    quit;

    %if %upcase(&dim_nm) = MANAGEMENT_ORG or %upcase(&dim_nm) = LEGAL_ORG %then         
      %let dim_column =organization_nm;        
    %else %let dim_column = &dim_nm._NM; 
         
    proc sql;
      create table &dim_nm._l as
      select 
        %if &prefix = %str() %then
        %do;
          &dim_nm._rk
        %end;
        %else
        %do;
          &dim_nm._rk as &prefix.&dim_nm._rk
        %end;
        %if %upcase(&dim_nm) = MANAGEMENT_ORG or %upcase(&dim_nm) = LEGAL_ORG %then
        %do;
          ,case
             when B.business_object_nm="&dim_nm" and 
                  A.&dim_nm._rk=B.business_object_rk and
                  upcase(B.language_cd) = "&localecd" then B.name_txt
             else A.organization_nm
           end as organization_nm len=%eval(200 * &char_multiplier)
        %end;
        %else
        %do;
          ,case
             when B.business_object_nm="&dim_nm" and 
                  A.&dim_nm._rk=B.business_object_rk and
                  upcase(B.language_cd) = "&localecd" then B.name_txt
            else A.&dim_nm._NM
            end as &dim_nm._NM len=%eval(200 * &char_multiplier)
        %end;
        from &src_lib..&dim_nm._l A left join
             &src_lib..full_sbo_trans B
        on B.business_object_nm="&dim_nm" and 
           A.&dim_nm._rk=B.business_object_rk and
           upcase(B.language_cd)="&localecd";
    quit;

    data &dim_nm.fmtin(keep=fmtname type start label);
      length fmtname $32 type $1 start $64;
      retain fmtname "&prefix.&dim_nm.fmt" type 'N';
      set &dim_nm._l;
        
      %if &prefix = %str() %then
      %do;
        start = strip(&dim_nm._rk);
      %end;
      %else
      %do;
        start = strip(&prefix.&dim_nm._rk);
      %end;
      %if %upcase(&dim_nm) = MANAGEMENT_ORG or %upcase(&dim_nm) = LEGAL_ORG %then
      %do;
        label = strip(organization_nm);
      %end;
      %else
      %do;
        label = strip(&dim_nm._nm);
      %end;
    run;

    proc format cntlin=&dim_nm.fmtin library=work;
    quit;
    run; 
      
    proc datasets nolist;
      delete &dim_nm._l &dim_nm.fmtin;
    quit;
    run;

    %put NOTE: {END: &sysmacroname};
	
  %mend rskmgt_create_dim_path_format;
  
  /* Format all dimensions */        
  %do I=1 %to &DIMCOUNT.;  
    %rskmgt_create_dim_path_format(dim_nm=&&DIMENSION&I);
  %end;  
  
  /* Change dimension rks to dimension names */ 
  /* instead of using formats since too many formats are not */
  /* handled well by VA */
  data &indsn;
    set &indsn;
    %do J=1 %to &DIMCOUNT; 
      length &prefix.&&DIMENSION&J.._path $%eval(4000 * &char_multiplier); 
	  
	  %rskmgt_varexist(dsn=&indsn,var=&prefix.&&DIMENSION&J.._rk_1,out_var=dim_node_exist);
     
     %if &dim_node_exist=1 %then
     %do;
       format &prefix.&&DIMENSION&J.._rk &prefix.&&DIMENSION&J.._rk_1 - &prefix.&&DIMENSION&J.._rk_&&&&&&DIMENSION&J.._lvls;
	  %end;
	  %else
	  %do;
	    format &prefix.&&DIMENSION&J.._rk;
	  %end; 
         
      %do k=1 %to &&&&&&DIMENSION&J.._lvls;  
        
        %if &k=1 %then
        %do;
        
		    %rskmgt_varexist(dsn=&indsn,var=&prefix.&&DIMENSION&J.._rk_1,out_var=dim_node_exist);
          %if &dim_node_exist=1 %then
          %do;
            if &prefix.&&DIMENSION&J.._rk ne . and &prefix.&&DIMENSION&J.._rk_1 = . then
            do;
              &prefix.&&DIMENSION&J.._Path=trim(left(put(&prefix.&&DIMENSION&J.._rk, &prefix.&&DIMENSION&J..fmt.)));
            end;
            else if &prefix.&&DIMENSION&J.._rk_1 ne . then 
            do;
              &prefix.&&DIMENSION&J.._Path=trim(left(put(&prefix.&&DIMENSION&J.._rk_1, &prefix.&&DIMENSION&J..fmt.)));
  
              %if &prefix ne  %then
              %do;
                label &prefix.&&DIMENSION&J.._Path="&prefix_label &&DIM_ALIAS&J %sysfunc(sasmsg(sashelp.rskmgtsvr,Path, NOQUOTE))";
              %end;
              %else
              %do;
                label &&DIMENSION&J.._Path="&&DIM_ALIAS&J %sysfunc(sasmsg(sashelp.rskmgtsvr,Path, NOQUOTE))";
              %end;
            end;
		    %end;
		    %else
		    %do;
		      if &prefix.&&DIMENSION&J.._rk ne . then
            do;
              &prefix.&&DIMENSION&J.._Path=trim(left(put(&prefix.&&DIMENSION&J.._rk, &prefix.&&DIMENSION&J..fmt.)));
            end;
	       %end;
        %end;
        %else
        %do;
          if &prefix.&&DIMENSION&J.._rk_&k ne . then 
          do;
            &prefix.&&DIMENSION&J.._Path=trim(left(&prefix.&&DIMENSION&J.._Path)) || ' > ' ||  trim(left(put(&prefix.&&DIMENSION&J.._rk_&k, &prefix.&&DIMENSION&J..fmt.))); 
              
            %if &prefix ne  %then
            %do;
              label &prefix.&&DIMENSION&J.._Path="&prefix_label &&DIM_ALIAS&J %sysfunc(sasmsg(sashelp.rskmgtsvr,Path, NOQUOTE))";
            %end;
            %else
            %do;
              label &prefix.&&DIMENSION&J.._Path="&&DIM_ALIAS&J %sysfunc(sasmsg(sashelp.rskmgtsvr,Path, NOQUOTE))";
            %end;
          end;
        %end;
      %end; 
    %end;   
       
    %if &prefix = %str() %then
    %do;
      %do r=1 %to &DIMCOUNT; 
        if &&DIMENSION&r.._rk ne . then
        do;
          length un_&&DIMENSION&r.._rk $%eval(200 * &char_multiplier);  
          un_&&DIMENSION&r.._rk=trim(left(put(&&DIMENSION&r.._rk, &&DIMENSION&r..fmt.))); 
          drop &&DIMENSION&r.._rk;
          label un_&&DIMENSION&r.._rk="&&DIM_ALIAS&r %sysfunc(sasmsg(sashelp.rskmgtsvr,dim_level_nm, NOQUOTE))";
          rename un_&&DIMENSION&r.._rk=&&DIMENSION&r.._rk;
        end;
         
        %do s=1 %to &&&&&&DIMENSION&r.._lvls;
          %rskmgt_varexist(dsn=&indsn,var=&&DIMENSION&r.._rk_&s,out_var=dim_node_exist);
          %if &dim_node_exist=1 %then
          %do;		  
		      if &&DIMENSION&r.._rk_&s ne . then
			   do;          
			     length un_&&&DIMENSION&r.._rk_&s $%eval(200 * &char_multiplier);      
			     un_&&DIMENSION&r.._rk_&s=trim(left(put(&&DIMENSION&r.._rk_&s, &&DIMENSION&r..fmt.))); 
			     drop &&DIMENSION&r.._rk_&s;
			     label un_&&DIMENSION&r.._rk_&s="&&DIM_ALIAS&r %sysfunc(sasmsg(sashelp.rskmgtsvr,dim_level_nm, NOQUOTE)) &s";
			     rename un_&&DIMENSION&r.._rk_&s=&&DIMENSION&r.._rk_&s; 
			   end;
          %end;			
        %end;
      %end;
    %end;
    %else
    %do;
      %do r=1 %to &DIMCOUNT;   
        drop &prefix.&&DIMENSION&r.._rk; 
        %do s=1 %to &&&&&&DIMENSION&r.._lvls;              
          drop &prefix.&&&DIMENSION&r.._rk_&s;
        %end;
      %end;
    %end;
  run;
  

%put NOTE: {END: &sysmacroname}; 

%exit: %mend rskmgt_apply_dim_path;