/******************************************************************/
/* Copyright (c) 2018 by SAS Institute Inc., Cary, NC, USA.       */
/* All Rights Reserved.                                           */
/******************************************************************/
/* NAME:     rskmgt_create_feature_dimensions.sas                 */
/* VERSION:  7.4                                                  */
/* DESCRIPTION:  Creates a table with a list of feature names,    */
/*               dimension names and dimension levels.            */
/* PRODUCT:  SAS Risk Management Server                           */
/******************************************************************/

%macro rskmgt_create_feature_dimensions;

%put NOTE: {START: &sysmacroname}; 
 
  %let inputFile=%trim(%sysfunc(pathname(work,l))/dimensionality.xml);

  /* Assign the xml file to a temporary work location */
  filename svc_rsp "&inputFile";
  
  /* Pull the dimensionality and feature information from the webdav server */
  proc http 
     out=svc_rsp 
     url="&srv_url.&rgfservice_name./rest/&contentId/systemConfig/dimensionalityXml" 
     method="get" 
     ct="application/x-www-form-urlencoded" 
     HTTP_TOKENAUTH;
  run;
 
  /* Dimension and Feature map used to read the dimensionality.xml file */
  filename dimmap "&rgf_control_file_path/dim_feature.map";
  /* xml libname used to read the xml file */
  libname svc_rsp xmlv2 xmlmap=dimmap access=readonly;
  
  /* Copy the xml files to the work area */
  proc copy in=svc_rsp out=work;
  run;

  /* Clear the filename statements */
  filename svc_rsp;
  filename dimmap;

  /* Create a dataset of the feature and dimensions */
  data feature_dims(keep=feature_nm dimension_nm);
    set feature dimension;
    by feature_ordinal;
    retain feature_nm;
 
    if not missing(feature_name) then feature_nm=feature_name;
    else feature_name=feature_nm;

    drop feature_name;

    if missing(dimension_name) then delete;

    rename dimension_name=dimension_nm;
  run;

  /* Create a dataset with the dimension alias names */
  data dim_alias;
    set alias;
    dimension_alias_nm=alias_new;
    dimension_nm=alias_old;

    keep dimension_alias_nm dimension_nm;
  run;
  
  /* Sort the data to merge the information into one table */
  proc sort data=dim_alias;
    by dimension_nm;
  run;

  proc sort data=feature_dims;
    by dimension_nm;
  run;

  
  data features_dimensions;
    merge feature_dims(in=a) dim_alias;
    by dimension_nm;
    if a;
  run; 
  /************************************************************************/

  
  data rgfcntl.dimensions;
    infile CNTLD(dimensions.txt) dlm=',' firstobs=2;
    length dimension_nm $32;
    length short_prefix $8;
    input dimension_nm short_prefix levels;
  run;
  
  data rgfcntl.dimension_loader_names;
    infile CNTLD(dimension_loader_names.txt) dlm=',' firstobs=2;
    length dimensionNm $32;
    length dimensionId_Nm $64;
    length dimensionSSCd_Nm $64;
    input dimensionNm dimensionId_Nm dimensionSSCd_Nm;
  run;

  /* Store the features, dimensions (original & alias names) and dimension levels */
  proc sql;
    create table feature_dims as
    select a.feature_nm, 
           case
             when a.dimension_nm=b.dimension_alias_nm then b.dimension_nm
             else a.dimension_nm
           end as dimension_nm,
          case
            when a.dimension_nm=b.dimension_alias_nm then b.dimension_alias_nm
            else ''
          end as dimension_alias_nm
    from features_dimensions a left join
         dim_alias b
    on a.dimension_nm=b.dimension_nm or
       a.dimension_nm=b.dimension_alias_nm
    order by a.feature_nm, dimension_nm, dimension_alias_nm;

    create table feature_dims_bo_nms as
    select 
      case
        when upcase(a.feature_nm)=upcase(b.business_object_type_nm) then upcase(b.business_object_nm)
        else upcase(a.feature_nm)
      end as feature_nm,
      case
        when upcase(a.dimension_nm)=upcase(c.business_object_type_nm) then upcase(c.business_object_nm)
        else a.dimension_nm
      end as dimension_nm,
      upcase(a.dimension_alias_nm) as dimension_alias_nm
    from feature_dims a left join 
         &src_lib..business_object_type b 
    on upcase(a.feature_nm)=upcase(b.business_object_type_nm)
    left join &src_lib..business_object_type c
    on upcase(a.dimension_nm)=upcase(c.business_object_type_nm)
    order by feature_nm, dimension_nm, dimension_alias_nm;

    create table &contentId.ctl.feature_dimensions as
    select case
             when a.feature_nm = 'CAUSE' then 'CAUSE_INST'
             when a.feature_nm = 'CONTROL' then 'CONTROL_INST'
             when a.feature_nm = 'INDICATOR' then 'KRI'
             when a.feature_nm = 'OBJECTIVE' then 'OBJECTIVE_INST'
             else a.feature_nm
           end as feature_nm,
           case
             when a.dimension_alias_nm <> a.dimension_nm then a.dimension_nm
             else ''
           end as dimension_nm,
           case
             when a.dimension_alias_nm ne '' then a.dimension_alias_nm
             else a.dimension_nm
           end as dimension_alias_nm,
           b.levels,
           c.dimensionId_Nm,
           c.dimensionSSCd_Nm
    from feature_dims_bo_nms a left join
         rgfcntl.dimensions b
    on a.dimension_nm=b.dimension_nm left join
         rgfcntl.dimension_loader_names c
    on a.dimension_nm=c.dimensionNm
    order by dimension_nm, feature_nm;
  quit;
   
  proc datasets nolist;
    delete features_dimensions feature_dims feature_dims_bo_nms dim_alias;
  run;
  quit;
  
  %global &contentId._features &contentId._dimensions &contentId._levels &contentId._distinct_dims &contentId._distinct_dims_levels;
     
  proc sql noprint;
    select feature_nm
             into :&contentId._features separated by '*'
    from &contentId.ctl.feature_dimensions;
        
    select dimension_nm
             into :&contentId._dimensions separated by '*'
    from &contentId.ctl.feature_dimensions;
      
    select levels
             into :&contentId._levels separated by '*'
    from &contentId.ctl.feature_dimensions;
  
    select distinct dimension_nm, levels 
             into : &contentId._distinct_dims separated by '*',
                  : &contentId._distinct_dims_levels separated by '*'
    from &contentId.ctl.feature_dimensions;
quit;
  
%put NOTE: {END: &sysmacroname}; 

%mend rskmgt_create_feature_dimensions;
