/************************************************************************************/
/*  Copyright (c) 2018 by SAS Institute, Inc., Cary, NC, USA.                       */
/*  All rights reserved.                                                            */
/************************************************************************************/
/*  NAME:        rskmgt_get_cust_field_labels.sas                                   */
/*  VERSION:     7.4                                                                */
/*  DESCRIPTION: Build a table (&contentId._Custommessages) of default and custom   */
/*               field labels.                                                      */
/*  PARAMETERS:                                                                     */
/*  PRODUCT:    SAS Risk Management Server                                          */
/************************************************************************************/
%macro rskmgt_get_cust_field_labels;

%put NOTE: {START: &sysmacroname};

  /* Get the custom field label from the sashelp.rskmgtsvr table */
  %if %sysfunc(exist(sashelp.rskmgtsvr)) %then
  %do;
    data rskmgtsvr(drop=lineno);
      length locale $5 key $200 text $1800;
      set sashelp.rskmgtsvr;
      where upcase(locale)=upcase("&localecd") or
            upcase(locale)="%substr(%upcase(&localecd),1,2)";
      locale=%upcase("&localecd");
    run;
    
    /* Sort the rskmgtsvr labels */
    proc sort data=rskmgtsvr out=custlbls;
      by key;
    run;
  %end;
  
  /* Get the custom field label from the sashelp.&contentId.srv table */
  /* If, &contentId is installed */
  %if %sysfunc(exist(sashelp.&contentId.srv)) %then
  %do;
    data &contentId.srv(drop=lineno);
      length locale $5 key $200 text $1800;
      set sashelp.&contentId.srv;
      where upcase(locale)=upcase("&localecd") or
            upcase(locale)="%substr(%upcase(&localecd),1,2)";
      locale=%upcase("&localecd");
    run;
    
    /* Sort the &contentId.srv labels */
    proc sort data=&contentId.srv out=custlbls;
      by key;
    run;
  %end;
  
  %if %sysfunc(exist(sashelp.rskmgtsvr)) and %sysfunc(exist(sashelp.&contentId.srv)) %then
  %do;
    data custlbls;
      set rskmgtsvr &contentId.srv;
    run;
    
    /* Sort the custlbls labels */
    proc sort data=custlbls;
      by key;
    run;
  %end;
  
  /* Create a table to store the custom messages for each &contentId */
  %local z contentId iLimit customMessages_cnt fullurl restLocale;
  
  %let restLocale=%trim(%sysfunc(getoption(locale)));
  
  %if %length(&restLocale) > 2 %then %let restLocale=%substr(%lowcase(&restLocale),1,2)_%substr(%upcase(&restLocale),4);
  %if %length(&restLocale) = 2 %then %let restLocale=%substr(%lowcase(&restLocale));
  
  %let z=1;
  %let contentId=%trim(%qscan(%bquote(&rgf_content_ids),&z,%str(,)));
  %do %while (&contentId ne );
  
    %if %sysfunc(exist(&contentId._Custommessages)) %then
    %do;
      proc sql;
        drop table &contentId._Custommessages;
      quit;
    %end;
   
    /* Construct the URL for the rest call */
    %let fullurl=&srv_url.&rgfservice_name./rest/&contentId/systemConfig/customMessages?locale=&restLocale%str(&)start=0%str(&)limit=10000;
      
    /* Call the rest service to pull the custom messages */
    filename svc_rsp temp;
    filename headout temp;
    filename jmap temp;
      
    proc http
      url= "&fullurl."
      method="GET"
      out=svc_rsp
      headerout=headout
      HTTP_TOKENAUTH ;
    run;
      
    /* JSON libname to the SAS datasets produced from the rest call */
    libname posts JSON fileref=svc_rsp map=jmap automap=create;
      
    /* Copy all JSON response datasets to work */
    proc copy in=posts out=work;
      select root items;
    run;
    
    /* Get the total # of customMessage records available */
    proc sql noprint;
      select count into : customMessages_cnt
      from work.root;
    quit;

    /* Initialize the Do Loop counter */
    %let iLimit=;
    
    /* Loop to pull all custom messages */ 
    %do %until (&iLimit >= &customMessages_cnt);   
      /* Check to see if there are any records in the items table */
      /* If so, continue; otherwise, exit the macro */
      %rskmgt_ds_exists(indsn=work.items,
                        output_nobs=items_cnt);
                    
      %if &items_cnt <= 0 %then %goto NoCustomMsgs;
      
      /* If work.items exist, merge w/ custObj_master */
      %if &items_cnt > 0 %then
      %do;
        /* Create a table to store the name of the custom messages */
        data work.customMsgs(keep=locale key text);
          set items;
          locale="%substr(%upcase(&localecd),1,2)";
          text=value;
        run;
         
        /* Now merge in the locale specific labels */
        proc sort tagsort data=work.customMsgs;
          by key;
        run;
      
        /* Create the master table structure to store the custom messages */
        data work.customMsgs_Master;
          attrib locale length=$%eval(5 * &char_multiplier) key length=$%eval(200 * &char_multiplier) text length=$%eval(1800 * &char_multiplier);
          stop;
        run;
    
        data work.customMsgs_Master;
          set work.customMsgs_Master work.customMsgs;
        run;
        
        /* Append customMsgs_Master to the &contentId._Custommessages dataset */
        proc append base=&contentId._customMsgs data=work.customMsgs_Master force;
        run;
                
        /* Remove the work datasets before looping again */
        proc datasets lib=work nolist;
          delete items root customMsgs_Master customMsgs;
        quit;
        run;
      %end;
           
      %let iLimit=%eval(&iLimit+10000);
     
      libname posts clear;
     
      %if &iLimit>=&customMessages_cnt %then %goto NoCustomMsgs;
    
      %let fullurl=&srv_url.&rgfservice_name./rest/&contentId/systemConfig/customMessages?locale=&restLocale%str(&)start=&iLimit%str(&)limit=10000; 

      proc http
        url="&fullurl."
        method="GET"
        out=svc_rsp
        headerout=headout
        HTTP_TOKENAUTH;
      run;

      libname posts JSON fileref=svc_rsp map=jmap automap=create;
   
      /* Copy all JSON response datasets to work */
      proc copy in=posts out=work;
        select root items;
      run;
     
      %NoCustomMsgs:
    %end; /* End - Loop to pull all custom message records */
  
   /* Sort the locale specific labels */
   proc sort tagsort data=work.&contentId._customMsgs;
     by key;
   run;
       
   /* Merge the sashelp.rskmgtsvr labels and locale specific labels */       
   data &contentId._Custommessages;
     merge custlbls &contentId._customMsgs;
     by key;
   run;
      
   /* Change all double quotes in the labels to single quotes */
   /* This is needed in order to add these labels to the custom field variables */
   data &contentId._Custommessages;
     set &contentId._Custommessages(rename=(key=orig_key));
     key=prxchange('s/(^.*)\.field\.(.*)\.displayName\.txt\s*$/$1_OBJ_$2/i',-1,orig_key);
     text=tranwrd(text,'"',"'");
     drop orig_key;
   run;
        
   proc sort data=&contentId._Custommessages nodupkey;
     by key;
   run;
   
   %let z=%eval(&z+1);
   %let contentId=%trim(%qscan(%bquote(&rgf_content_ids),&z,%str(,))); 
  %end; /* End - Loop to pull custom messages by contentId */
 
%put NOTE: {END: &sysmacroname};

%mend rskmgt_get_cust_field_labels;