/**********************************************************************/
/*  Copyright (c) 2018 by SAS Institute, Inc., Cary, NC, USA.         */
/*  All rights reserved.                                              */
/**********************************************************************/
/*  NAME:        rskmgt_job_env_setup.sas                             */
/*  VERSION:     7.4                                                  */
/*  DESCRIPTION: Setup the environment for all jobs used to generate  */
/*               data used Visual Analytics.                          */
/*  PARAMETERS:                                                       */
/*  PRODUCT:     SAS Risk Management Server                           */
/**********************************************************************/

%macro rskmgt_job_env_setup;

%put NOTE: {START: &sysmacroname}; 
    
  /* Create a metadata table to store the name of the custom object */
  /* table alias names for each &contentId */
  %local z contentId iLimit custObj_cnt fullurl;
  
  %let z=1;
  %let contentId=%trim(%qscan(%bquote(&rgf_content_ids),&z,%str(,)));
  %do %while (&contentId ne );
  
    %if %sysfunc(exist(&contentId.ctl.custObj_tbl_ref)) %then
    %do;
      proc sql;
        drop table &contentId.ctl.custObj_tbl_ref;
      quit;
    %end;
   
    /* Construct the URL for the rest call */
    %let fullurl=&srv_url.&rgfservice_name./rest/&contentId/systemConfig/customObjects?enabled=true%str(&)sortBy=customObjectIndex%str(&)start=0%str(&)limit=&&&contentId._maxLimit;
      
    /* Call the rest service to pull the custom objects list */
    filename svc_rsp temp;
    filename headout temp;
    filename jmap temp;
      
    proc http
      url= "&fullurl."
      method="GET"
      out=svc_rsp
      headerout=headout
      HTTP_TOKENAUTH ;
    run;
      
    /* JSON libname to the SAS datasets produced from the rest call */
    libname posts JSON fileref=svc_rsp map=jmap automap=create;
      
    /* Copy all JSON response datasets to work */
    proc copy in=posts out=work;
      select root items;
    run;
    
    /* Get the total # of customObject records available */
    proc sql noprint;
      select count into : custObj_cnt
      from work.root;
    quit;

    /* Initialize the Do Loop counter */
    %let iLimit=;
    
    /* Loop to pull all custom objects */ 
    %do %until (&iLimit >= &custObj_cnt);   
      /* Check to see if there are any records in the items table */
      /* If so, continue; otherwise, exit the macro */
      %rskmgt_ds_exists(indsn=work.items,
                        output_nobs=items_cnt);
                    
      %if &items_cnt <= 0 %then %goto NoCustomObjects;
      
      /* If work.items exist, merge w/ custObj_master */
      %if &items_cnt > 0 %then
      %do;
        /* Create a table to store the name of the custom objects */
        data work.custObjTbls;
          set work.items;
          cust_obj_tbl_no=customObjectIndex;
          cust_obj_tbl_nm='cust_obj_' || strip(cust_obj_tbl_no) || '_l';
          cust_obj_alias_nm=alias;
          if enabled=1 then featureEnabled='true';
          else featureEnabled='false';
        
          j=0;
          x=0;
          do until(j=0);
            if alias ne '' then j=notalnum(trim(alias), j+1);

            if j>0 then
            do;
              newvar=compress(tranwrd(alias,substr(alias, j, 1),'_'));
              if j>0 and x>1 then newvar=compress(tranwrd(newvar,substr(alias, j, 1),'_'));
              x=x+1;
            end;
          end;
          if j=0 and x=0 then cust_obj_alias_nm=alias;
          else cust_obj_alias_nm=newvar;
      
          keep cust_obj_tbl_no cust_obj_tbl_nm cust_obj_alias_nm featureEnabled;
        run;
         
        proc sort tagsort data=work.custObjTbls;
          by cust_obj_tbl_no;
        run;
           
        /* Create the master table structure to store the custom objects */
        data custObj_master;
          attrib cust_obj_tbl_no length=8. cust_obj_tbl_nm length=$%eval(32 * &char_multiplier) cust_obj_alias_nm length=$%eval(32 * &char_multiplier)
                 featureEnabled length=$%eval(5 * &char_multiplier);
          stop;
        run;
      
        data work.custObj_master;
          set work.custObj_master work.custObjTbls;
        run;
        
        /* Append the &contentId.ctl.custObj_tbl_ref dataset to a master dataset */
        proc append base=&contentId.ctl.custObj_tbl_ref data=work.custObj_master force;
        run;
                
        /* Remove the work datasets before looping again */
        proc datasets lib=work nolist;
          delete items root custObjTbls custObj_master;
        quit;
        run;
      %end;
           
     %let iLimit=%eval(&iLimit+&&&contentId._maxLimit);
     
     libname posts clear;
     
     %if &iLimit>=&custObj_cnt %then %goto NoCustomObjects;
    
     %let fullurl=&srv_url.&rgfservice_name./rest/&contentId/systemConfig/customObjects?enabled=true%str(&)sortBy=customObjectIndex?start=&iLimit%str(&)limit=&&&contentId._maxLimit; 

     proc http
       url="&fullurl."
       method="GET"
       out=svc_rsp
       headerout=headout
       HTTP_TOKENAUTH;
     run;

     libname posts JSON fileref=svc_rsp map=jmap automap=create;
   
     /* Copy all JSON response datasets to work */
     proc copy in=posts out=work;
     run;
     
     %NoCustomObjects:
   %end; /* End - Loop to pull all customObjectsTblAlias records */
    
    /* If the buildETL_props dataset exist, execute this section of code to identify which data sources should/should not be built */
    %if %sysfunc(exist(work.&contentId._buildETL_props)) %then
    %do;
      data work.&contentId._etl_props;
        length cust_obj_tbl_no 8. cust_obj_tbl_nm $%eval(32 * &char_multiplier) buildETLData loadETLData $%eval(5 * &char_multiplier);
        set &contentId._buildETL_props;
        key_nm=scan(key,-2,'.');
        cust_obj_tbl_no=substr(key_nm,anydigit(key_nm),length(key_nm)-length(anydigit(key_nm)));
        cust_obj_tbl_nm='cust_obj_' || substr(key_nm,anydigit(key_nm),length(key_nm)-length(anydigit(key_nm))) || '_l';
        if findw(key,'BUILDETLDATA')>0 then buildETLData=lowcase(value);
        if findw(key,'LOADETLDATA')>0 then loadETLData=lowcase(value);
        /* Turn off the load of the ETL if the build is turned off */
        if buildETLData='false' or buildETLData='no' then loadETLData='false';
        keep cust_obj_tbl_no cust_obj_tbl_nm buildETLData loadETLData;
      run;

      proc sort tagsort data=work.&contentId._etl_props;
        by cust_obj_tbl_no;
      run;
     
	  data work.&contentId._etl_props;
        length _buildETLData _loadETLData $%eval(5 * &char_multiplier);
        set work.&contentId._etl_props;
        by cust_obj_tbl_no;
        retain _buildETLData _loadETLData;
        if first.cust_obj_tbl_no then
        do;
          _buildETLData=buildETLData;
          _loadETLData=loadETLData;
        end;
        else 
        do;
          if _buildETLData='' then _buildETLData=buildETLData;
          if _loadETLData='' then _loadETLData=loadETLData;
        end;
        if last.cust_obj_tbl_no then 
        do;
          drop buildETLData loadETLData;
          rename _buildETLData=buildETLData _loadETLData=loadETLData;
          output;
        end;
      run;
	  
      data &contentId.ctl.custObj_tbl_ref;
        merge &contentId.ctl.custObj_tbl_ref work.&contentId._etl_props;
        by cust_obj_tbl_no;
        if buildETLData='' then buildETLData='true';
        if loadETLData='' then loadETLData='true';
      run;
    %end; /* End - Loop to pull custom objects */
  
    %let z=%eval(&z+1);
    %let contentId=%trim(%qscan(%bquote(&rgf_content_ids),&z,%str(,))); 
  %end; /* End - Loop to pull custom objects by contentId */
  
  /* Get the locale from the SAS server */
  %global localecd;
  %let localecd=%trim(%upcase(%sysfunc(getoption(locale))));
  
  /* Load &intblnm into LASR (Yes or No) */
  /* Reload LASR permissions on data sources (Yes or No) */
  %global load_lasr reload_perm;

  %rskmgt_get_app_property_value(application=&RGF_SWC_MID_CLASSID, 
								         property=grc.loadLASRTbls,
								         attribute=ClassIdentifier,
                                 output_property_value=load_lasr_out, 
                                 output_status=rgf_status);

  %let load_lasr=%trim(%nrstr(&load_lasr_out));
  
  %rskmgt_get_app_property_value(application=&RGF_SWC_MID_CLASSID, 
								         property=grc.reloadLASRPermissions,
								         attribute=ClassIdentifier,
                                 output_property_value=reload_perm_out, 
                                 output_status=rgf_status);

  %let reload_perm=%trim(%nrstr(&reload_perm_out));
  
  /* Connect to LASR if &RGF_CONN_LASR = YES */
  /* and load the link instance table. */
  %if %upcase(&RGF_CONN_LASR) = YES %then
  %do;
    /* Setup the LASR environment variables and */
    /* connect to the LASR library */
    %rskmgt_get_lasr_env;
  %end;
  
  %let z=1;
  %let contentId=%trim(%qscan(%bquote(&rgf_content_ids),&z,%str(,)));
  %do %while (&contentId ne );
    /* Sets a variable used to hold the &contentId.DB libref */
    %global src_lib;
    %let src_lib=&&&contentId._src_lib; 
       
    /* Load the &contentId.ctl features and dimension tables. */
    %rskmgt_create_feature_dimensions;
  
    /* Build the permission table used for data filtering */
    %rskmgt_create_security_denorm(srclib=&src_lib,
                                   dstlib=work);

    /* LASR permission filter macros */
    %rskmgt_lasr_perm_filter; 
    
    /* Load the &contentId._src_lib.link instance table. */
    %if %upcase(&RGF_CONN_LASR) = YES %then
    %do;
      /* Rebuild the &contentId.dat.&contentId._link_instance tables */ 
      %rskmgt_load_link_inst_tbl; 
    %end;
  
    /* Build the &contentId._Custommessages dataset w/ custom object labels */
    %rskmgt_get_cust_field_labels;
  
    %let z=%eval(&z+1);
    %let contentId=%trim(%qscan(%bquote(&rgf_content_ids),&z,%str(,))); 
  %end;

%put NOTE: {END: &sysmacroname}; 

%mend rskmgt_job_env_setup;
