/**********************************************************************/
/*  Copyright (c) 2019 by SAS Institute, Inc., Cary, NC, USA.         */
/*  All rights reserved.                                              */
/**********************************************************************/
/*  NAME:        rskmgt_add_linked_users.sas                          */
/*  VERSION:     7.4                                                  */
/*  DESCRIPTION: Adds the linked users to the output dataset.         */
/*  PRODUCT:     SAS Risk Management Server                           */
/**********************************************************************/

%macro rskmgt_add_linked_users;

%put NOTE: {START: &sysmacroname};

/************************/
/* Add the linked users */
/************************/
/* Subset the links by &business_object_type_nm and user objects */
proc sql;
  create table work.all_user_links as
  select *
  from &contentId.dat.&contentId._link_instances_l
  where ((upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and
         upcase(business_object_type_nm_2)="USER")
     or (upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
         upcase(business_object_type_nm_1)="USER"));
quit;

/* Create User Display Name Column to use in the transpose below */
proc sql;
  create table work.filtered_links_users as
  select A.*
         ,B.display_nm as user_nm length=256
         ,case
           when business_object_type_nm_1='user' then scan(link_type_nm,2,"|")
           when business_object_type_nm_2='user' then scan(link_type_nm,1,"|")
         else ''
         end as user_lbl length=256
  from work.all_user_links A left join
       &src_lib..orauser_l B
  on ((a.business_object_type_nm_1='user' and
       a.business_object_rk_1=b.user_rk ) or
      (a.business_object_type_nm_2='user' and 
       a.business_object_rk_2=b.user_rk))
  order by link_type_rk, link_instance_rk;
quit;

/* Check to see if the work.filtered_links_users dataset exist */
/* and if user links exist */
%rskmgt_ds_exists(indsn=work.filtered_links_users,
                  output_nobs=filter_user_links_cnt);

%if &filter_user_links_cnt = 0 %then %goto exit;

%if %symexist(&contentId._concat_user_links) %then
%do;
  %if %upcase(&&&contentId._concat_user_links)=YES or %upcase(&&&contentId._concat_user_links)=TRUE %then
  %do;
    /* Transpose the link_type_nm for the Business Object 1 User link_types only */
    proc transpose data=work.filtered_links_users(where=(business_object_type_nm_1='user'))
      out=work.bo1_user_links(where=(link_type_rk ^=.) drop=_NAME_ _LABEL_) let;
      var user_nm;
      copy business_object_rk_2 link_type_id user_lbl;
      id link_type_id;
      by link_type_rk link_instance_rk;
      idlabel user_lbl;
    run;
    quit;

    /* Sort the datasets before merging together */
    proc sort tagsort data=work.bo1_user_links(rename=(business_object_rk_2=&rk_column));
      by &rk_column link_type_rk;
    run;

    /* Transpose the link_type_nm for the Business Object 2 User link_types only */
    proc transpose data=work.filtered_links_users(where=(business_object_type_nm_2='user'))
      out=work.bo2_user_links(where=(link_type_rk ^=.) drop=_NAME_ _LABEL_) let;
      var user_nm;
      copy business_object_rk_1 link_type_id user_lbl;
      id link_type_id;
      by link_type_rk link_instance_rk;
      idlabel user_lbl;
    run;
    quit;

    /* Sort the datasets before merging together */
    proc sort tagsort data=work.bo2_user_links(rename=(business_object_rk_1=&rk_column));
      by &rk_column link_type_rk;
    run;

    /* Put the BO1 and BO2 datasets together */
    data work.user_links;
      merge work.bo1_user_links work.bo2_user_links;
      by &rk_column link_type_rk;
    run;

    /* Sort the user_links dataset to aggregate the multi-users */
    proc sort tagsort data=work.user_links;
      by &rk_column link_type_rk;
    run;

    /* Create a list of user link types w/ more than one value */
    proc sql noprint;
	  create table work.single_user_lnks as
	  select link_type_rk, link_type_id, user_lbl, count(distinct(business_object_rk_1)) as link_cnt
	  from work.filtered_links_users
	  where business_object_type_nm_1='user'
	  group by link_type_id
	  having count(distinct(business_object_rk_1))=1
	  union
	  select link_type_rk, link_type_id, user_lbl, count(distinct(business_object_rk_2)) as link_cnt
	  from work.filtered_links_users
	  where business_object_type_nm_2='user'
	  group by link_type_id
	  having count(distinct(business_object_rk_2))=1;
	  
      create table work.multi_user_lnks as
      select link_type_rk, link_type_id, user_lbl, count(distinct(business_object_rk_1)) as link_cnt
      from work.filtered_links_users
      where business_object_type_nm_1='user'
      group by link_type_id
      having count(distinct(business_object_rk_1))>1
      union
      select link_type_rk, link_type_id, user_lbl, count(distinct(business_object_rk_2)) as link_cnt
      from work.filtered_links_users
      where business_object_type_nm_2='user'
      group by link_type_id
      having count(distinct(business_object_rk_2))>1;
    quit;

    /* Check to see if the multi_user_lnks dataset exist */
    %rskmgt_ds_exists(indsn=work.multi_user_lnks,
                      output_nobs=multi_links_cnt);

    /* Expand the link_types columns w/ more than 1 value and roll up the data */
    %if &multi_links_cnt > 0 %then
    %do;
      /* Put the list of user links into macro variable lists */
      %let multi_links=;
      %let multi_link_num=;
      %let multi_usr_lbl=;
     
      proc sql noprint;
        select link_type_id, 
               link_cnt,
               user_lbl  into : multi_links separated by ',',
                              : multi_link_num separated by ',',
                              : multi_usr_lbl separated by ','
        from work.multi_user_lnks;
      quit;
     
      %local ma mb mc md me mf mg multi_link_nm
             multi_link_nm2 multi_link_nm3 multi_link_nm4
             multi_link_nm5 multi_link_nm6 multi_link_var
             multi_link_lbl multi_link_cnt;  
      %let ma=1;
      %let mb=1;
      %let mc=1;
      %let md=1;
      %let me=1;
      %let mf=1;
      %let mg=1; 
      %let multi_link_nm=%scan(%bquote(&multi_links),&ma,%str(","));
      %let multi_link_cnt=%scan(%bquote(&multi_link_num),&ma,%str(","));
      %let multi_link_nm2=%scan(%bquote(&multi_links),&mb,%str(","));
      %let multi_link_nm3=%scan(%bquote(&multi_links),&mc,%str(","));
      %let multi_link_nm4=%scan(%bquote(&multi_links),&md,%str(","));
      %let multi_link_nm5=%scan(%bquote(&multi_links),&me,%str(","));
      %let multi_link_nm6=%scan(%bquote(&multi_links),&mf,%str(","));
      %let multi_link_var=%scan(%bquote(&multi_links),&mg,%str(","));
      %let multi_link_lbl=%scan(%bquote(&multi_usr_lbl),&mg,%str(","));
     
      data work.user_links;
        %do %while (&multi_link_nm ne );
          %if &ma=1 %then
          %do;
            length
          %end; 

          &multi_link_nm _&multi_link_nm $%eval((256 * &multi_link_cnt) + (&multi_link_cnt-1))

          %let ma=%eval(&ma+1);
          %let multi_link_nm=%scan(%bquote(&multi_links),&ma,%str(","));
          %let multi_link_cnt=%scan(%bquote(&multi_link_num),&ma,%str(","));
        %end;
        %str(;)

        %do %while (&multi_link_nm2 ne );
          %if &mb=1 %then
          %do;
            retain
          %end;

          _&multi_link_nm2

          %let mb=%eval(&mb+1);
          %let multi_link_nm2=%scan(%bquote(&multi_links),&mb,%str(","));
        %end;
        %str(;)

        set work.user_links;
        by &rk_column link_type_rk;
        if first.link_type_rk then
        do;
          %do %while (&multi_link_nm3 ne );
            _&multi_link_nm3=cats(&multi_link_nm3);

            %let mc=%eval(&mc+1);
            %let multi_link_nm3=%scan(%bquote(&multi_links),&mc,%str(","));
          %end;
        end;
        else
        do;
          %do %while (&multi_link_nm4 ne );
            _&multi_link_nm4=catx("&&&contentId._user_delimiter",_&multi_link_nm4,&multi_link_nm4);

            %let md=%eval(&md+1);
            %let multi_link_nm4=%scan(%bquote(&multi_links),&md,%str(","));
          %end;
        end;

        if last.link_type_rk then output;
       
        drop
     
        %do %while (&multi_link_nm5 ne );
          &multi_link_nm5

          %let me=%eval(&me+1);
          %let multi_link_nm5=%scan(%bquote(&multi_links),&me,%str(","));
        %end;
     
        %str(;)
     
        rename
         
        %do %while (&multi_link_nm6 ne );
          _&multi_link_nm6=&multi_link_nm6

          %let mf=%eval(&mf+1);
          %let multi_link_nm6=%scan(%bquote(&multi_links),&mf,%str(","));
        %end;
     
        %str(;)

       label
      
       %do %while (&multi_link_var ne );
         _&multi_link_var="&multi_link_lbl"

         %let mg=%eval(&mg+1);
         %let multi_link_var=%scan(%bquote(&multi_links),&mg,%str(","));
         %let multi_link_lbl=%scan(%bquote(&multi_usr_lbl),&mg,%str(","));
       %end;
     
       %str(;)
      run;   

      %let ma=;
      %let mb=;
      %let mc=;
      %let md=;
      %let me=;
      %let mf=;
      %let mg=;
      %let multi_link_nm=;
      %let multi_link_nm2=;
      %let multi_link_nm3=;
      %let multi_link_nm4=;
      %let multi_link_nm5=;
      %let multi_link_nm6=;
      %let multi_link_var=;
      %let multi_link_lbl=;
      %let multi_link_cnt=;
	  
	  /* Check to see if the single_user_lnks dataset exist */
      %rskmgt_ds_exists(indsn=work.single_user_lnks,
                        output_nobs=single_links_cnt);
		
	  data work.all_user_lnks;
		set work.multi_user_lnks 
		  
		%if &single_links_cnt > 0 %then
        %do;
		  work.single_user_lnks
		%end;
		%str(;)
	  run;
    %end;  
	
	/* Sort the work.user_links */
	proc sort tagsort data=work.user_links(drop=link_type_rk link_type_id user_lbl link_instance_rk);
	  by &rk_column;
	run;
    
    /* Get a distinct list of user link_type_ids */
    %let usr_links=;
    %let usr_link_lbl=;
	%let usr_link_len_lst=;
    %let usr_link_num=;
	%let usr_link_rk=;
    		
	%if &multi_links_cnt = 0 and &single_links_cnt > 0 %then
    %do;
	  data work.all_user_lnks;
        set work.single_user_lnks;
	  run;
	%end;
	
    /* Check to see if the all_user_lnks dataset exist */
    %rskmgt_ds_exists(indsn=work.all_user_lnks,
                      output_nobs=all_links_cnt);
    
	%if &all_links_cnt > 0 %then
    %do;
	  proc sql noprint;
		select distinct link_type_rk, link_type_id, user_lbl, link_cnt, 
		 	   'link_'||strip(put(link_type_rk,8.))||'_len' as usr_link_len into : usr_link_rk separated by ',',
		 																		 : usr_links separated by ',',
																				 : usr_link_lbl separated by ',',
																				 : usr_link_num separated by ',',
																				 : usr_link_len_lst separated by ' '
		from work.all_user_lnks
		order by link_type_rk;
	  quit;    
	%end;

	  %local ua ub uc ud ue uf ug uh uh1 uh2 uh3 uh4 user_link_nm
	         user_link_nm1 user_link_nm2 user_link_nm3
		     user_link_nm4 user_link_nm5 user_link_var user_link_lbl
		     user_link_cnt user_link_len_lst user_link_len_lst1
		     user_link_len_lst2 user_link_len_lst3 user_link_len_lst4;
	  %let ua=1;
	  %let ub=1;
	  %let uc=1;
	  %let ud=1;
	  %let ue=1;
	  %let uf=1;
	  %let ug=1;
	  %let uh=1;
	  %let uh1=1;
	  %let uh2=1;
	  %let uh3=1;
	  %let uh4=1;

	  %let user_link_nm=%scan(%bquote(&usr_links),&ua,%str(","));
	  %let user_link_nm1=%scan(%bquote(&usr_links),&ub,%str(","));
	  %let user_link_nm2=%scan(%bquote(&usr_links),&uc,%str(","));
	  %let user_link_nm3=%scan(%bquote(&usr_links),&ud,%str(","));
	  %let user_link_nm4=%scan(%bquote(&usr_links),&uf,%str(","));
	  %let user_link_nm5=%scan(%bquote(&usr_links),&ue,%str(","));
	  %let user_link_var=%scan(%bquote(&usr_links),&ug,%str(","));
	  %let user_link_lbl=%scan(%bquote(&usr_link_lbl),&ug,%str(","));
	  %let user_link_cnt=%scan(%bquote(&usr_link_num),&ua,%str(","));
	  %let user_link_len_lst=%scan(%bquote(&usr_link_len_lst),&uh,%str(" "));
	  %let user_link_len_lst1=%scan(%bquote(&usr_link_len_lst),&uh1,%str(" "));
	  %let user_link_len_lst2=%scan(%bquote(&usr_link_len_lst),&uh2,%str(" "));
	  %let user_link_len_lst3=%scan(%bquote(&usr_link_len_lst),&uh3,%str(" "));
	  %let user_link_len_lst4=%scan(%bquote(&usr_link_len_lst),&uh4,%str(" "));
		 
	  /* Roll up the data by &rk_column */
	  data work.user_links;	 
	    %do %while (&user_link_nm ne );
		  %if &ua=1 %then %str(length);
		  
	      _&user_link_nm $%eval((32 * &user_link_cnt) + (&user_link_cnt-1))
		  _&user_link_len_lst 8.

		  %let ua=%eval(&ua+1);
		  %let uh=%eval(&uh+1);
		  %let user_link_nm=%scan(%bquote(&usr_links),&ua,%str(","));
		  %let user_link_cnt=%scan(%bquote(&usr_link_num),&ua,%str(","));
		  %let user_link_len_lst=%scan(%bquote(&usr_link_len_lst),&uh,%str(" "));
		  
		  %if &user_link_nm = %then %str(;);
	    %end;

	    set user_links;
	    by &rk_column;
	    where &rk_column ne .;

	      %do %while (&user_link_nm1 ne );
		    %if &ub=1 %then %str(retain);
	        _&user_link_nm1
		    _&user_link_len_lst1

		    %let ub=%eval(&ub+1);
		    %let uh1=%eval(&uh1+1);
		    %let user_link_nm1=%scan(%bquote(&usr_links),&ub,%str(","));
	        %let user_link_len_lst1=%scan(%bquote(&usr_link_len_lst),&uh1,%str(" "));
			
			%if &user_link_nm1 = %then %str(;);
		  %end;
		  %str(;)

	    if first.&rk_column then 
	    do;
		  %do %while (&user_link_nm2 ne );
		    _&user_link_nm2=&user_link_nm2;
		    _&user_link_len_lst2=length(_&user_link_nm2);

		    %let uc=%eval(&uc+1);
		    %let uh2=%eval(&uh2+1);
		    %let user_link_nm2=%scan(%bquote(&usr_links),&uc,%str(","));
		    %let user_link_len_lst2=%scan(%bquote(&usr_link_len_lst),&uh2,%str(" "));
		  %end;
	    end;
	    else
	    do;
		  %do %while (&user_link_nm3 ne );
		    _&user_link_nm3=catx('',_&user_link_nm3,&user_link_nm3);
		    _&user_link_len_lst3=length(_&user_link_nm3);

		    %let ud=%eval(&ud+1);
		    %let uh3=%eval(&uh3+1);
		    %let user_link_nm3=%scan(%bquote(&usr_links),&ud,%str(","));
		    %let user_link_len_lst3=%scan(%bquote(&usr_link_len_lst),&uh3,%str(" "));
		  %end;
	    end;
		 
	    if last.&rk_column then output;

	      %do %while (&user_link_nm5 ne );
		    %if &ue=1 %then %str(drop);
			
		    &user_link_nm5

		    %let ue=%eval(&ue+1);
		    %let user_link_nm5=%scan(%bquote(&usr_links),&ue,%str(","));
			
			%if &user_link_nm5 = %then %str(;);
		  %end;
		 
		  %do %while (&user_link_nm4 ne );
		    %if &uf=1 %then %str(rename);
			
	 	    _&user_link_nm4=&user_link_nm4
		    _&user_link_len_lst4=&user_link_len_lst4

		    %let uf=%eval(&uf+1);
		    %let uh4=%eval(&uh4+1);
		    %let user_link_nm4=%scan(%bquote(&usr_links),&uf,%str(","));
		    %let user_link_len_lst4=%scan(%bquote(&usr_link_len_lst),&uh4,%str(" "));
			
			%if &user_link_nm4 = %then %str(;);
		  %end;

	      %do %while (&user_link_var ne );
		    %if &ug=1 %then %str(label);
			
		    _&user_link_var="&user_link_lbl"

		    %let ug=%eval(&ug+1);
		    %let user_link_var=%scan(%bquote(&usr_links),&ug,%str(","));
		    %let user_link_lbl=%scan(%bquote(&usr_link_lbl),&ug,%str(","));
			
			%if &user_link_var = %then %str(;);
		  %end;
	  run;   

	  %let ua=;
	  %let ub=;
	  %let uc=;
	  %let ud=;
	  %let ue=;
	  %let uf=;
	  %let ug=;
	  %let uh=;
	  %let uh1=;
	  %let uh2=;
	  %let uh3=;
	  %let uh4=;
	  %let user_link_nm=;
	  %let user_link_nm1=;
	  %let user_link_nm2=;
	  %let user_link_nm3=;
	  %let user_link_nm4=;
	  %let user_link_nm5=;
	  %let user_link_var=;
	  %let user_link_lbl=;
	  %let user_link_cnt=;
	  %let user_link_len_lst=;
	  %let user_link_len_lst1=;
	  %let user_link_len_lst2=;
	  %let user_link_len_lst3=;
	  %let user_link_len_lst4=;
		
	%if &all_links_cnt > 0 %then
    %do;
	  /* Get the max length for each link_type_id in order to shrink the column widths */
	  /* to a size that has been adjusted to the max length of each user */
	  proc means data=work.user_links noprint;
	    var &usr_link_len_lst;
	    output out=work.user_max_len(drop=_TYPE_ _FREQ_) max=;
	  run;

	  proc transpose data=work.user_max_len 
        	         out=work.trans_user_lens(rename=(col1=user_link_max_len)) name=user_link_type;
	  run;
			
	  proc sql noprint;
	    select user_link_type, user_link_max_len  into : user_link_nm_lst separated by ',',
		 											   : user_data_len_lst separated by ','
	    from work.trans_user_lens
	    order by user_link_type;
	  quit;
		
	  /* Modify the work.user_links table to scope the user link column lengths down to actual data storage size */
	  /* needed instead of estimated size based on total number of customObject links by link type */
	  %local user_link_val user_var_len user_link_len user_len user_link_var;
	  %let user_var_len=1;
	  %let user_link_len=1;
	  %let user_link_val=;
	  %let user_len=;
	  %let user_link_var=;
		
	  %let user_link_var=%scan(%bquote(&usr_links),&user_var_len,%str(","));
	  %let user_link_val=%scan(%bquote(&user_link_nm_lst),&user_var_len,%str(","));
	  %let user_len=%scan(%bquote(&user_data_len_lst),&user_var_len,%str(","));

	  proc sql;
	    alter table work.user_links
		  
	    modify
		  %do %while (&user_link_var ne );
		    &user_link_var. char(&user_len.)

		    %let user_var_len=%eval(&user_var_len+1);
		    %let user_link_var=%scan(%bquote(&usr_links),&user_var_len,%str(","));
		    %let user_len=%scan(%bquote(&user_data_len_lst),&user_var_len,%str(","));
			
		    %if &user_link_var ne %then %str(,);
		  %end;
		  
	    drop
	      %do %while (&user_link_val ne );
		    &user_link_val.

		    %let user_link_len=%eval(&user_link_len+1);
		    %let user_link_val=%scan(%bquote(&user_link_nm_lst),&user_link_len,%str(","));
			
		    %if &user_link_val ne %then %str(,);
		  %end;
		  %str(;)  
	  quit;
		
	  %let user_var_len=;
	  %let user_link_len=;
	  %let user_link_val=;
	  %let user_len=;
	  %let user_link_var=;
	%end;
  %end;
  /* Do not load the concatenated User links */
  %else %if %upcase(&&&contentId._concat_user_links)=NO or %upcase(&&&contentId._concat_user_links)=FALSE %then
  %do;
    /* Transpose the link_type_nm for the USER link_types only */
    proc transpose data=work.filtered_links_users(where=(business_object_type_nm_1='user' or business_object_type_nm_2='user'))
     out=work.user_lnks(drop=_NAME_ _LABEL_) suffix=_1 let;
      var user_nm;
      copy business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2;
      id link_type_id;
      by link_type_rk link_instance_rk;
     idlabel user_lbl;
    run;
    quit;
    
    /* Create the &rk_column to use in merging back with the rest of the link instance data */
    data work.user_links;
      length &rk_column 8.;
      set work.user_lnks;
      if business_object_type_nm_2='user' then &rk_column=business_object_rk_1;
      else if business_object_type_nm_1='user' then &rk_column=business_object_rk_2;
      if &rk_column ne . then output;
    run;
    
    /* Sort the work.user_links to prep for a merge */
    proc sort tagsort data=work.user_links;
      by &rk_column;
    run;
  %end;
%end;   

%put NOTE: {END: &sysmacroname};

%exit: %mend rskmgt_add_linked_users;
