/**********************************************************************/
/*  Copyright (c) 2019 by SAS Institute, Inc., Cary, NC, USA.         */
/*  All rights reserved.                                              */
/**********************************************************************/
/*  NAME:        rskmgt_filter_link_inst_tbl.sas                      */
/*  VERSION:     7.4                                                  */
/*  DESCRIPTION: Filter the link instance table for a given BO TYPE.  */
/*  PARAMETERS:  business_object_type_nm = business object type name  */
/*                                         used to filter the list of */
/*                                         link instances             */
/*               indsn = primary business object table that will be   */
/*                       joined to the 360 links                      */
/*               rk_column = key used to join the &indsn dataset      */
/*                           to the business object link instances    */
/*  PRODUCT:     SAS Risk Management Server                           */
/**********************************************************************/

%macro rskmgt_filter_link_inst_tbl(business_object_type_nm=,
                                   indsn=,
                                   rk_column=);

%put NOTE: {START: &sysmacroname};

%if %symexist(&contentId._load_links) %then
%do;
  %if %upcase(&&&contentId._load_links)=YES or %upcase(&&&contentId._load_links)=TRUE %then
  %do; 
    /* Subset the filtered_link by &business_object_type_nm */
    proc sql;
      create table work.filtered_links as
      select *
      from &contentId.dat.&contentId._link_instances_l
      where upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" or
            upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)";
    quit;

    /* Stop processing if the work.filtered_links dataset does not exist */
    /* or is empty */
    %local nobs rc;
    %let dsid = %sysfunc(open(work.filtered_links));
    %if &dsid %then 
    %do;
      %let nobs  = %sysfunc(attrn(&dsid, nlobs));
      %let rc = %sysfunc(close(&dsid));
    %end;
    %else %let nobs  = 0;
     
    %if &nobs > 0 %then 
    %do;
      /* Sort the work.filtered_links by business_object_rk_1 first */
      proc sort tagsort data=work.filtered_links;
        by link_type_rk link_instance_rk;
      run;
     
      /* Remove the 'RK' from the business_object_rk_nm1 fields in order to add the suffix below and */
      /* to not go over the 32 character variable limit */
      /* Filter out the linked users - they will be processed separately */
      data work.filtered_links_no_users;
        set work.filtered_links;
        where business_object_type_nm_1^='user' and business_object_type_nm_2^='user';
        business_object_rk_nm1=substr(trim(business_object_rk_nm1),1,length(trim(business_object_rk_nm1))-2);
        business_object_rk_nm2=substr(trim(business_object_rk_nm2),1,length(trim(business_object_rk_nm2))-2);
      run;

      /* Transpose the business_object_rk_1 column */
      proc transpose data=work.filtered_links_no_users out=work.bo1_links(drop=_NAME_ _LABEL_) suffix=B1 let;
        var business_object_rk_1;
        copy business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2;
        id business_object_rk_nm1;
        by link_type_rk link_instance_rk;
        idlabel business_object_type_nm_1;
      run;
      quit;

      /* Rename the variables to drop the suffix for all GLB* variables except the &rk_field.B1 variable */
      %let rk_field=;
     
      proc sql noprint;
        select distinct business_object_rk_nm1 into: rk_field trimmed
        from work.filtered_links_no_users
        where upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)";
      quit;
     
      %let num_bo1_vars=;
      %let B1_rk_field_exist=;
     
      proc sql noprint;
        select count(name) into : num_bo1_vars
        from dictionary.columns
        where upcase(libname)="WORK" and upcase(memname)="BO1_LINKS" and  
              upcase(name) ne "&rk_field.B1" and substr(upcase(name),1,3)='GLB';
      
        select count(*) as B1_rk_field into : B1_rk_field_exist
        from dictionary.columns
        where upcase(libname)="WORK" and upcase(memname)="BO1_LINKS" and
              upcase(name)="&rk_field.B1";
       quit;
    
       %if &num_bo1_vars > 0 %then
       %do;
         proc sql noprint;
           select distinct(name), substr(trim(name),1,length(name)-2) into : var_bo1_1-:var_bo1_%trim(%left(&num_bo1_vars)),
                                                                           : newvar_bo1_1-:newvar_bo1_%trim(%left(&num_bo1_vars))
           from dictionary.columns
           where upcase(libname)="WORK" and upcase(memname)="BO1_LINKS" and
                 upcase(name) ne "&rk_field.B1" and substr(upcase(name),1,3)='GLB';
         quit;
      
         proc datasets library = WORK nolist;
           modify bo1_links;
           rename
             %do i = 1 %to &num_bo1_vars.;
               &&var_bo1_&i = &&newvar_bo1_&i..RK
             %end;
           ;
         quit;
         run;
       %end;
     
       /* Sort the work.filtered_links by business_object_rk_2 */
       proc sort tagsort data=work.filtered_links_no_users;
         by link_type_rk link_instance_rk;
       run;

       /* Transpose the business_object_rk_2 column */
       proc transpose data=work.filtered_links_no_users out=work.bo2_links(drop=_NAME_ _LABEL_) suffix=B2 let;
         var business_object_rk_2;
         copy business_object_rk_1 business_object_type_nm_1  business_object_rk_2 business_object_type_nm_2;
         id business_object_rk_nm2;
         by link_type_rk link_instance_rk;
         idlabel business_object_type_nm_2;
       run;
       quit;

       /* Rename the variables to drop the suffix for all GLB* variables except the &rk_field.B2 variable */
       %let num_bo2_vars=;
       %let B2_rk_field_exist=;
     
       proc sql noprint;
         select count(name) into : num_bo2_vars
         from dictionary.columns
         where upcase(libname)="WORK" and upcase(memname)="BO2_LINKS" and  
               upcase(name) ne "&rk_field.B2" and substr(upcase(name),1,3)='GLB';
             
         select count(*) as B2_rk_field into : B2_rk_field_exist
         from dictionary.columns
         where upcase(libname)="WORK" and upcase(memname)="BO2_LINKS" and
               upcase(name)="&rk_field.B2";
       quit;
     
       %if &num_bo2_vars > 0 %then
       %do;
         proc sql noprint;
           select distinct(name), substr(trim(name),1,length(name)-2) into : var_bo2_1-:var_bo2_%trim(%left(&num_bo2_vars)),
                                                                           : newvar_bo2_1-:newvar_bo2_%trim(%left(&num_bo2_vars))
           from dictionary.columns
           where upcase(libname)="WORK" and upcase(memname)="BO2_LINKS" and
                 upcase(name) ne "&rk_field.B2" and substr(upcase(name),1,3)='GLB';
         quit;
      
         proc datasets library = WORK nolist;
           modify bo2_links;
           rename
             %do h = 1 %to &num_bo2_vars.;
               &&var_bo2_&h = &&newvar_bo2_&h..RK
             %end;
            ;
         quit;
         run;
       %end;

       /* Sort the work.filtered_links_no_users by link_type_rk */
       proc sort tagsort data=work.filtered_links_no_users;
         by link_type_rk link_instance_rk;
       run;

       /* Transpose the link_type_nm column */
       proc transpose data=work.filtered_links_no_users out=work.botype_links(drop=_NAME_ _LABEL_) let;
         var link_type_nm;
         copy business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2;
         id link_type_id;
         by link_type_rk link_instance_rk;
         idlabel link_type_nm;
       run;
       quit;
             
       /* Sort the work.filtered_links_no_users by link_type_id */
       proc sort tagsort data=work.filtered_links_no_users;
         by link_type_id;
       run;

       /* Sort the work.bo2_links to prep for a merge */
       proc sort data=work.bo1_links;
         by link_type_rk link_instance_rk business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2;
       run;

       /* Sort the work.bo2_links to prep for a merge */
       proc sort tagsort data=work.bo2_links;
         by link_type_rk link_instance_rk business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2;
       run;

       /* Update the bo1 and bo2 tables */
       data work.bo12_links;
         update work.bo1_links work.bo2_links;
         by link_type_rk link_instance_rk business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2;
       run;

       /* Sort the work.botype_links to prep for a merge */
       proc sort tagsort data=work.botype_links;
         by link_type_rk link_instance_rk business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2;
       run;

       /* Merge in the bo1 & bo2 links with the botypes */
       data work.bo_links;
         merge bo12_links botype_links;
         by link_type_rk link_instance_rk business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2;
       run;
       
       /* Get a list of all link_type_ids, link_type_nms and related rk columns */
       /* in order to build a data structure to merge with the &indsn table */
       /* Check to see if the rk columns have been renamed and pull names from */
       /* &contentId._Custommessages to use for the rk colum labels.  If not, use the rk column */
       /* default object names for the labels */
       %let vars=;
       %let vars_val=;
       %let rk_keys=;
       %let rk_lbls=;

       proc sql noprint; 
         select distinct (link_type_id) as link_type_id into :vars separated by " "
         from work.filtered_links_no_users
         order by link_type_id;

         select distinct ("'"|| trim(link_type_nm) ||"'") as link_type_nm, link_type_id into :vars_val separated by "*",
                                                                                             : dummy separated by "*"
         from work.filtered_links_no_users
         order by link_type_id;

         create table rk_key_lbls(drop=name) as
         select distinct (substr(substr(name,5),1,length(substr(name,5))-3)) as rk_key, (upcase(label)) as key
                ,name
         from sashelp.vcolumn
         where libname='WORK' and 
               memname='BO_LINKS' and
               substr(name,1,4)='GLB_'
         order by name;

         create table cust_key_lbls as
         select distinct key as key_nm, scan(key,2,'.') as key, text as label
         from work.&contentId._custommessages
         where substr(upcase(key),1,24 )='BUSINESSOBJECTDESCRIPTOR';

         create table cust_rk_lbls as
         select a.rk_key, b.key_nm, 
                case
                  when b.label ne '' then b.label
                  else propcase(a.key)
                end as label, 
                a.key
         from rk_key_lbls a left join
              cust_key_lbls b
         on a.key=upcase(b.key);

         select rk_key, label into : rk_keys separated by '*',
                                   : rk_lbls separated by '*'
         from cust_rk_lbls;
       quit;

       %let i=1;
       %let j=1;
       %let h=1;
       %let var_nm=%qscan(%bquote(&vars),&i,' ');
       %let var_val=%qscan(%bquote(&vars_val),&i,'*');
       %let rk_key_nm=%scan(&rk_keys,&j,'*');
       %let rk_key_lbl=%qscan(%bquote(&rk_lbls),&j,'*');
     
       /* Parse &rk_column in order to get the base object name */
       /* in order to create the &rk_key_base_nm._count variable */
       %if %sysfunc(substr(%upcase(&business_object_type_nm),1,12)) eq CUSTOMOBJECT %then
       %do;
         /* Get the rk_key name for the current business_object_type_nm */
         %let rk_key_base_nm=;

         proc sql noprint;
           select distinct rk_key into : rk_key_base_nm trimmed
           from cust_rk_lbls
           where upcase(key)="%upcase(&business_object_type_nm)";
         quit;

         %if &rk_key_base_nm = %str() %then %let rk_key_base_nm=%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3));
       %end;
       %else
       %do;
         %let rk_key_base_nm=%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3));
       %end;
       
       /* Join the &indsn, work.bo_links and link_types */
       proc sort tagsort data=&indsn;
         by &rk_column;
       run;
    
       proc sort tagsort data=work.bo_links out=work.bo_links1(rename=(business_object_rk_1=&rk_column)
                                            where=(upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)"));
         by business_object_rk_1;
       run;
    
       data merge_bolinks1;
         merge &indsn (in=master)
               work.bo_links1 (in=bo_links1);
         by &rk_column;
         if master;
      
         %sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3))_key=strip(put(&rk_column,8.));
         label &rk_column="%sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Rk, NOQUOTE))"
               %sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3))_key="%sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Key, NOQUOTE))"
         ;

         %if &B1_rk_field_exist=1 or &B2_rk_field_exist=1 %then
         %do;
           %if &B1_rk_field_exist=1 %then
           %do;
             if upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
                &rk_column ^= GLB_&rk_key_base_nm._B1 then GLB_&rk_key_base_nm._RK=GLB_&rk_key_base_nm._B1;
           %end;

           %if &B2_rk_field_exist=1 %then
           %do;
             if upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
                &rk_column ^= GLB_&rk_key_base_nm._B2 then GLB_&rk_key_base_nm._RK=GLB_&rk_key_base_nm._B2;
           %end;

           label GLB_&rk_key_base_nm._RK="%sysfunc(sasmsg(sashelp.rskmgtsvr,Linked, NOQUOTE)) %sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Rk, NOQUOTE))";
         %end;

         %if &B1_rk_field_exist=1 or &B2_rk_field_exist=1 %then
         %do;
           %if &B1_rk_field_exist=1 %then
           %do;
             if upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
                &rk_column ^= GLB_&rk_key_base_nm._B1 then GLB_&rk_key_base_nm._Key=strip(put(GLB_&rk_key_base_nm._B1,8.));
           %end;

           %if &B2_rk_field_exist=1 %then
           %do;
             if upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
                &rk_column ^= GLB_&rk_key_base_nm._B2 then GLB_&rk_key_base_nm._Key=strip(put(GLB_&rk_key_base_nm._B2,8.));
           %end;
         
           label GLB_&rk_key_base_nm._Key="%sysfunc(sasmsg(sashelp.rskmgtsvr,Linked, NOQUOTE)) %sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Global_Key, NOQUOTE))";
         %end;

         %do %while (&rk_key_nm ne );
           label GLB_&rk_key_nm._RK="%sysfunc(sasmsg(sashelp.rskmgtsvr,Linked, NOQUOTE)) &rk_key_lbl %sysfunc(sasmsg(sashelp.rskmgtsvr,Global_Rk, NOQUOTE))";

           GLB_&rk_key_nm._Key=strip(put(GLB_&rk_key_nm._RK,8.));
           label GLB_&rk_key_nm._Key="%sysfunc(sasmsg(sashelp.rskmgtsvr,Linked, NOQUOTE)) &rk_key_lbl %sysfunc(sasmsg(sashelp.rskmgtsvr,Global_Key, NOQUOTE))";

           %let j=%eval(&j+1);
           %let rk_key_nm=%scan(&rk_keys,&j,'*');
           %let rk_key_lbl=%qscan(%bquote(&rk_lbls),&j,'*');
         %end;

         %do %while (&var_nm ne );
           label &var_nm="&var_val";

           %let i=%eval(&i+1);
           %let var_nm=%qscan(%bquote(&vars),&i,' ');
           %let var_val=%qscan(%bquote(&vars_val),&i,'*');
         %end;
       
       run;
    
       proc sort tagsort data=work.bo_links out=work.bo_links2(rename=(business_object_rk_2=&rk_column)
                                            where=(upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)"));
         by business_object_rk_2;
       run;
    
       %let j=1;
       %let rk_key_nm=%scan(&rk_keys,&j,'*');
       %let rk_key_lbl=%qscan(%bquote(&rk_lbls),&j,'*');
     
       data merge_bolinks2;
         merge &indsn (in=master)
               work.bo_links2 (in=bo_links2 where=(&rk_column ne .));
         by &rk_column;
         if bo_links2;
       
         %sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3))_key=strip(put(&rk_column,8.));
         label &rk_column="%sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Rk, NOQUOTE))"
               %sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3))_key="%sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Key, NOQUOTE))"
         ;

         %if &B1_rk_field_exist=1 or &B2_rk_field_exist=1 %then
         %do;
           %if &B1_rk_field_exist=1 %then
           %do;
             if upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
                &rk_column ^= GLB_&rk_key_base_nm._B1 then GLB_&rk_key_base_nm._RK=GLB_&rk_key_base_nm._B1;
           %end;

           %if &B2_rk_field_exist=1 %then
           %do;
             if upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
                &rk_column ^= GLB_&rk_key_base_nm._B2 then GLB_&rk_key_base_nm._RK=GLB_&rk_key_base_nm._B2;
           %end;

           label GLB_&rk_key_base_nm._RK="%sysfunc(sasmsg(sashelp.rskmgtsvr,Linked, NOQUOTE)) %sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Rk, NOQUOTE))";
         %end;

         %if &B1_rk_field_exist=1 or &B2_rk_field_exist=1 %then
         %do;
           %if &B1_rk_field_exist=1 %then
           %do;
             if upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
                &rk_column ^= GLB_&rk_key_base_nm._B1 then GLB_&rk_key_base_nm._Key=strip(put(GLB_&rk_key_base_nm._B1,8.));
           %end;

           %if &B2_rk_field_exist=1 %then
           %do;
             if upcase(business_object_type_nm_1)="%upcase(&business_object_type_nm)" and upcase(business_object_type_nm_2)="%upcase(&business_object_type_nm)" and
                &rk_column ^= GLB_&rk_key_base_nm._B2 then GLB_&rk_key_base_nm._Key=strip(put(GLB_&rk_key_base_nm._B2,8.));
           %end;
         
           label GLB_&rk_key_base_nm._Key="%sysfunc(sasmsg(sashelp.rskmgtsvr,Linked, NOQUOTE)) %sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Global_Key, NOQUOTE))";
         %end;
        
         %do %while (&rk_key_nm ne );
           label GLB_&rk_key_nm._RK="%sysfunc(sasmsg(sashelp.rskmgtsvr,Linked, NOQUOTE)) &rk_key_lbl %sysfunc(sasmsg(sashelp.rskmgtsvr,Global_Rk, NOQUOTE))";

           GLB_&rk_key_nm._Key=strip(put(GLB_&rk_key_nm._RK,8.));
           label GLB_&rk_key_nm._Key="%sysfunc(sasmsg(sashelp.rskmgtsvr,Linked, NOQUOTE)) &rk_key_lbl %sysfunc(sasmsg(sashelp.rskmgtsvr,Global_Key, NOQUOTE))";

           %let j=%eval(&j+1);
           %let rk_key_nm=%scan(&rk_keys,&j,'*');
           %let rk_key_lbl=%qscan(%bquote(&rk_lbls),&j,'*');
         %end;
       run;
    
       data &business_object_type_nm._lnks;
         set merge_bolinks1 merge_bolinks2;
      
         drop link_type_rk link_instance_rk business_object_rk_1 business_object_type_nm_1 business_object_rk_2 business_object_type_nm_2
              GLB_&rk_key_base_nm._B1 GLB_&rk_key_base_nm._B2;
       run;
     
       proc sort tagsort data=&business_object_type_nm._lnks;
         by &rk_column;
       run;
       
       /* Create the User links dataset */
       %rskmgt_add_linked_users;
       
       %rskmgt_ds_exists(indsn=work.user_links,
                  output_nobs=user_links_cnt);

       %if &user_links_cnt > 0 %then
       %do;
         data work.&business_object_type_nm._lnks;
           merge work.&business_object_type_nm._lnks work.user_links;
           by &rk_column;
         run; 
       %end;       
          
       /* Initialize the &business_object_type_nm._count variable to count just the first.&business_object_type_nm._rk */
       proc sql noprint;
         create table work.rels as
         select *, monotonic() as rownum, min(calculated rownum) as min,
           case when calculated rownum=calculated min 
             then 1 
             else 0 
           end as %sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3))_count 
           length=8 
           format=NLNUM12.
           label="%sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Count, NOQUOTE))"
         from work.&business_object_type_nm._lnks
         group by &rk_column
         order by rownum;
       quit;

       /* Remove the rownum and min columns */
       proc sql noprint;
         alter table work.rels
         drop rownum, min;
       quit;
    
       /* Drop the &indsn and rename the rels table to &indsn */
       proc datasets library=work nolist;
         delete &indsn;
         change rels=&indsn;
       run;
       quit;
     
     %end;
     %else /* Add the count variable to the &indsn dataset */
     %do;
       /* Parse &rk_column in order to get the base object name */
       /* in order to create the &rk_key_base_nm._count variable */
       %if %sysfunc(substr(%upcase(&business_object_type_nm),1,12)) eq CUSTOMOBJECT %then
       %do;
         /* Get the rk_key name for the current business_object_type_nm */
         %let rk_key_base_nm=;

         %if &rk_key_base_nm = %str() %then %let rk_key_base_nm=%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3));
       %end;
       %else
       %do;
         %let rk_key_base_nm=%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3));
       %end;
      
       /* Create the User links dataset */
       %rskmgt_add_linked_users;
       
       %rskmgt_ds_exists(indsn=work.user_links,
                  output_nobs=user_links_cnt);

       %if &user_links_cnt > 0 %then
       %do;
         data work.&business_object_type_nm._lnks;
           merge work.&business_object_type_nm._lnks work.user_links;
           by &rk_column;
         run; 
       %end;  
       
       /* Initialize the &business_object_type_nm._count variable to count just the first.&business_object_type_nm._rk */
       proc sql;
         create table work.rels as
         select *, monotonic() as rownum, min(calculated rownum) as min,
         case when calculated rownum=calculated min 
              then 1 
              else 0 
         end as %sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3))_count 
         length=8 
         format=NLNUM12.
         label="%sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Count, NOQUOTE))"
        from work.&indsn
        group by &rk_column
        order by rownum;
       quit;
     
       /* Remove the rownum and min columns */
       proc sql noprint;
         alter table work.rels
         drop rownum, min;
       quit;
      
       /* Drop the &indsn and rename the rels table to &indsn */
       proc datasets library=work nolist;
         delete &indsn;
         change rels=&indsn;
       run;
       quit;
     %end;
  %end;
  /* Only load the User links (if they exist) */
  %else %if %upcase(&&&contentId._load_links)=NO or %upcase(&&&contentId._load_links)=FALSE %then
  %do;     
    /* Create the User links dataset */
    %rskmgt_add_linked_users;
    
    %rskmgt_ds_exists(indsn=work.user_links,
                      output_nobs=user_links_cnt);

    %if &user_links_cnt > 0 %then
    %do;    
      /* Merge the User links with the rest of the &business_object_type_nm._lnks */
      data work.&indsn;
        merge work.&indsn work.user_links;
        by &rk_column;
      run;  
    %end;
    
    /* Initialize the &business_object_type_nm._count variable to count just the first.&business_object_type_nm._rk */
    proc sql;
      create table work.rels as
      select *, monotonic() as rownum, min(calculated rownum) as min,
      case when calculated rownum=calculated min 
           then 1 
           else 0 
      end as %sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3))_count 
      length=8 
      format=NLNUM12.
      label="%sysfunc(propcase(%sysfunc(substr(&rk_column,1,%sysfunc(length(&rk_column))-3)))) %sysfunc(sasmsg(sashelp.rskmgtsvr,Count, NOQUOTE))"
      from work.&indsn
      group by &rk_column
      order by rownum;
    quit;
     
    /* Remove the rownum and min columns */
    proc sql noprint;
      alter table work.rels
      drop rownum, min;
    quit;
      
    /* Drop the &indsn and rename the rels table to &indsn */
    proc datasets library=work nolist;
      delete &indsn;
      change rels=&indsn;
    run;
    quit;
  %end;
%end;   
         
%put NOTE: {END: &sysmacroname};

%mend rskmgt_filter_link_inst_tbl;