/**********************************************************************/
/*  Copyright (c) 2018 by SAS Institute, Inc., Cary, NC, USA.         */
/*  All rights reserved.                                              */
/**********************************************************************/
/*  NAME:        rskmgt_job_audit_custobj_data.sas                    */
/*  VERSION:     7.4                                                  */
/*  DESCRIPTION: Create customObject audit data.                      */
/*  PARAMETERS:                                                       */
/*  PRODUCT:     SAS Model Risk Management                            */
/**********************************************************************/

%macro rskmgt_job_audit_custobj_data;

%put NOTE: {START: &sysmacroname};

/* Store the list of customObject numbers needed in a list */
  proc sql noprint;
    select cust_obj_tbl_no into : customObjectNums separated by ','
    from &contentId.ctl.custObj_tbl_ref;
  quit;

  /* Get a count of the number of customObjects that are enabled */
  %let customObjectList=%sysfunc(countw(%bquote(&customObjectNums)));

  /* Lookup the custom object table names in metadata to see if they have been renamed */
  %do z=1 %to &customObjectList;
  
    /* Get a value out of the customObjectList macro list */
    %let custObjValue=%trim(%qscan(%bquote(&customObjectNums),&z,%str(,)));
        
    /* Get the alias name, buildETLData and loadETLData properties */
    %let custObjAliasNm=;
    %let buildETLDataTbls=;
    %let loadETLDataTbls=;

    proc sql noprint;
      select cust_obj_alias_nm, buildETLData, loadETLData into : custObjAliasNm trimmed,
                                                               : buildETLDataTbls,
                                                               : loadETLDataTbls
      from &contentId.ctl.custobj_tbl_ref
      where cust_obj_tbl_nm="cust_obj_&custObjValue._l";
    quit;
    
    /* Store the output dataset name and rk column name in macros vars for */
    /* macro calls further downstream */
    %let outdsn=%trim(&contentId._&custObjAliasNm._A);
    %let out_rk_col=&custObjAliasNm._rk;

    /* If buildETLDataTbls=true, then proceed.  Otherwise, move to the next customObject */
    %if &buildETLDataTbls=false or &buildETLDataTbls=no %then %goto continue;

    %if %eval(%sysfunc(exist(&contentId.dat.&outdsn)) = 1) %then
    %do;
      proc datasets lib=&contentId.dat nolist;
        delete &outdsn;
      quit;
      run;
    %end;

    %if %eval(%sysfunc(exist(rgflasr.&outdsn)) = 1 ) %then
    %do;
      proc datasets lib=rgflasr nolist;
        delete &outdsn;
      quit;
      run;

      proc metalib;
        omr (liburi="SASLibrary?@libref='RGFLASR'");
        select ("&outdsn");
        update_rule=(delete);
      run;
    %end;

    %if %eval(%sysfunc(exist(rgfrload.&outdsn)) = 1 ) %then
    %do;
      proc datasets lib=rgfrload nolist;
        delete &outdsn;
      quit;
      run;
    %end;
 
    /* Get the custom object rest call names */
    %rskmgt_get_rest_obj_names(business_obj_type_nm=customObject&custObjValue);
   
    /* Get a complete list of all customObject keys available */
    proc sql;
     create table customObjectKeys as
     select distinct cust_obj_&custObjValue._rk
     from &src_lib..cust_obj_&custObjValue.
     order by cust_obj_&custObjValue._rk;
    quit;
     
    data customObjectKeys;
      set customObjectKeys;
      rowno=_n_;
    run;
     
    proc sql noprint;
      select count(*) into : totHistRecs
      from customObjectKeys;
    quit;
     
    %let iHistLimit=;
    %let maxHistLimit=100;
  
    /* Loop to pull all customObjects records */ 
    %do %until (&iHistLimit >= &totHistRecs);

      /* Put the keys into a list and subset by the maxHistLimit */
      %let custObjKeys=;
       
      proc sql noprint;
        select cust_obj_&custObjValue._rk into : custObjKeys separated by ','
        from work.customObjectKeys
       
        %if &iHistLimit = %str() %then
        %do;
          where rowno <= &maxHistLimit;
        %end;
        %else
        %do;
          where rowno between %eval(&iHistLimit+1) and %eval(&iHistLimit+&maxHistLimit);
        %end;
      quit;
      
      /* Check to see if there are any records in the business object table */
      /* If so, continue; otherwise, exit the macro */
      %rskmgt_ds_exists(indsn=customObjectKeys,
                        output_nobs=custObj_cnt);
                    
      %if &custObj_cnt = 0 %then %goto continue;
   
      /* Construct the URL for the rest call */
      %let fullurl=&srv_url.&rgfservice_name./rest/&contentId./history/&restobj.?filter=and(in(key,&custObjKeys),in(revisionTypes,FIELDS,LOCATION));
   
      /* Call the rest service to pull the business object custom fields and links */
      filename svc_rsp temp;
      filename headout temp;
      filename jmap temp;
   
      proc http
        url= "&fullurl."
        method="GET"
        out=svc_rsp
        headerout=headout
        HTTP_TOKENAUTH ;
      run;
   
      /* JSON libname to the SAS datasets produced from the rest call */
      libname posts JSON fileref=svc_rsp map=jmap automap=create;
   
      /* Copy all JSON response datasets to work */
      proc copy in=posts out=work;
      run;
      
      /* Check to see if there are any records in the customFields table */
      /* If so, continue; otherwise, exit the macro */
      %rskmgt_ds_exists(indsn=work.revisions_values,
                        output_nobs=revisions_values_cnt);
                    
      %if &revisions_values_cnt <= 0 %then %goto continue;
         
      /* Rename the custom object table and fields to use the alias names */
      /* Create data structure for the primary customObject table and use the alias names */
      proc sql;
        create table work.&custObjAliasNm
          (source_system_cd char(%eval(3 * &char_multiplier)),
           &custObjAliasNm._rk num,
           &custObjAliasNm._id char(%eval(32 * &char_multiplier)),
           &custObjAliasNm._nm char(%eval(1000 * &char_multiplier)),
           &custObjAliasNm._desc char(%eval(4000 * &char_multiplier)),
           created_user_name char(%eval(256 * &char_multiplier)),
           last_update_user_name char(%eval(256 * &char_multiplier)),
           created_Dttm num
		                  format=32.,
           valid_From_Dttm num
                         format=32.,
           valid_To_Dttm num
                         format=32.,
           editor_user_name char(%eval(256 * &char_multiplier)),
           change_reason char(%eval(4000 * &char_multiplier)), 
           status_dttm num
                         format=32.,
           created_from_rk num,
           customUser1 char(%eval(256 * &char_multiplier)),
           customUser2 char(%eval(256 * &char_multiplier)),
           customUser3 char(%eval(256 * &char_multiplier)),
           customUser4 char(%eval(256 * &char_multiplier)),
           customUser5 char(%eval(256 * &char_multiplier)));
      quit;
         
      /* Create a list of all variables that are available in the revision_values table */
      /* b/c not all of these fields are required and may not be in the returned rest call */
      %rskmgt_var_exists(libname=work,
                         memname=revisions_values,
                         varListNm=CustObjDefVars);
                            
      /* Create a list of all variables that are available in the revisions_reasons table */
      /* b/c not all of these fields are required and may not be in the returned rest call */
      %rskmgt_var_exists(libname=work,
                         memname=revisions_reason,
                         varListNm=changeReasonVars);
           
      /* Create the primary customObject table and use the alias names */
      proc sql;
        insert into work.&custObjAliasNm
        select A.sourceSystemCd as source_system_cd 
               ,A.key as &custObjAliasNm._rk
               ,A.objectid as &custObjAliasNm._id
                
               %if %index(&CustObjDefVars,NAME) > 0 %then
               %do;
                 ,A.name as &custObjAliasNm._nm
               %end;
               %else
               %do;
                 ,'' as &custObjAliasNm._nm
               %end;
                 
               %if %index(&CustObjDefVars,DESCRIPTION) > 0 %then
               %do;
                 ,A.description as &custObjAliasNm._desc
               %end;
               %else
               %do;
                 ,'' as &custObjAliasNm._desc
               %end;
                 
               ,CRE.title as created_user_name
               ,UPD.title as last_update_user_name
               ,case
                  when A.createdDttm ne '' then ((dhms(DATEPART(input(A.createdDttm,e8601dt.)), HOUR(input(A.createdDttm,e8601dt.)), MINUTE(input(A.createdDttm,e8601dt.)), second(input(A.createdDttm,e8601dt.)))- 315619200)*1000)
                  else .
                end as created_Dttm
               ,case
                  when B.validFrom ne '' then ((dhms(DATEPART(input(B.validFrom,e8601dt.)), HOUR(input(B.validFrom,e8601dt.)), MINUTE(input(B.validFrom,e8601dt.)), second(input(B.validFrom,e8601dt.)))- 315619200)*1000)
                  else .
                end as valid_From_Dttm
               ,case
                  when B.validTo ne '' then ((dhms(DATEPART(input(B.validTo,e8601dt.)), HOUR(input(B.validTo,e8601dt.)), MINUTE(input(B.validTo,e8601dt.)), second(input(B.validTo,e8601dt.)))- 315619200)*1000)
                  else .
                end as valid_To_Dttm
               ,CR.title as editor_user_name 
                
               %if %index(&changeReasonVars,REASONTXT) > 0 %then
               %do;
                 ,CR.reasonTxt as change_reason 
               %end;
               %else
               %do;
                 ,'' as change_reason
               %end;
                 
               ,case
                  when B.validFrom ne '' then ((dhms(DATEPART(input(B.validFrom,e8601dt.)), HOUR(input(B.validFrom,e8601dt.)), MINUTE(input(B.validFrom,e8601dt.)), second(input(B.validFrom,e8601dt.)))- 315619200)*1000)
                  else .
                end as status_dttm 
                 
               %if %index(&CustObjDefVars,CREATEDFROM) > 0 %then
               %do;
                 ,A.createdFrom as created_from_rk
               %end;
               %else
               %do;
                 ,. as created_from_rk
               %end;
                 
               ,'' as customUser1
               ,'' as customUser2
               ,'' as customUser3
               ,'' as customUser4
               ,'' as customUser5
                 
        from work.revisions_values A JOIN
             work.items_revisions B
        ON A.ordinal_revisions=B.ordinal_revisions
        LEFT JOIN
          (select
             title,
             ordinal_values
           from work.values_links
           where rel='creator'
          ) CRE
          ON A.ordinal_values=CRE.ordinal_values
          LEFT JOIN
            (select
              title,
              ordinal_values
             from work.values_links
             where rel='updater'
           ) UPD
           ON A.ordinal_values=UPD.ordinal_values
           LEFT JOIN
             (select
                RL.title,
                RR.reasonTxt,
                RR.ordinal_revisions
              from work.reason_links RL join
                   work.revisions_reason RR
              on RL.ordinal_reason=RR.ordinal_reason and
                 RL.rel='editor'
             ) CR
           ON A.ordinal_revisions=CR.ordinal_revisions
         ;
       quit;
   
       /* Check to see if there are any records in the business object table */
       /* If so, continue; otherwise, exit the macro */
       %rskmgt_ds_exists(indsn=&custObjAliasNm,
                         output_nobs=busObjObs_cnt);
   
       %if &busObjObs_cnt = 0 %then %goto continue;
       
     /* Add the custom fields */
     %rskmgt_add_hist_cust_fields(indsn=&custObjAliasNm,
                                  feature=CUST_OBJ_&custObjValue,
                                  rk_column=&out_rk_col,
                                  business_object_type_nm=customObject&custObjValue);
     
     %let iHistLimit=%eval(&iHistLimit+&maxHistLimit);
	 
	/* Fix the date and datetime formats */
    %rskmgt_modify_date_time(indsn=&custObjAliasNm);
	   
     /* Append the &outdsn dataset to a master dataset */
     proc append base=&custObjAliasNm._fnl data=&custObjAliasNm force;
     run;
     
     /* Check to see if there are any records in the customFields table */
     /* If so, continue; otherwise, exit the macro */
     %rskmgt_ds_exists(indsn=work.values_customfields,
                       output_nobs=custFieldVals);

     /* Remove the work datasets before looping again */
     proc datasets lib=work nolist;
       delete items_revisions
              %if &CustObjDefVars > 0 %then
              %do;
                revisions_values 
              %end;
              
              %if &changeReasonVars > 0 %then
              %do;
                revisions_reason
              %end;

	           %if &custFieldVals > 0 %then
              %do;
                values_customfields
              %end;		
              
              values_links reason_links;
     quit;
     run;
   %end; /* End - Loop to pull all customObjects records */
 
    /* Add active_flg to identify which customObjects are active vs. deleted */
    proc sort tagsort data=&custObjAliasNm._fnl;
      by &out_rk_col.;
    run;
  	  
    /* Add dimensions */
    %rskmgt_sp_unfold_dimensions(srclib=&src_lib,
                                 in=&custObjAliasNm._fnl,
                                 rk_column=&out_rk_col,
                                 feature=CUST_OBJ_&custObjValue,
                                 business_object_nm=CUST_OBJ_&custObjValue);

    /* Apply dimension path */
    %rskmgt_apply_dim_path(indsn=&custObjAliasNm._fnl,
                           feature=CUST_OBJ_&custObjValue);


    /* Sort the &outdsn._fnl*/
    proc sort tagsort data=work.&custObjAliasNm._fnl(drop=dim_point_rk);
      by &out_rk_col;
    run;
  
    /* Add custom object labels */
    %let custObjLabel=;
 
    /* Query the custommessages dataset to get the custom object labels */
    proc sql noprint;
      select text into : custObjLabel
      from work.&contentId._custommessages
      where upcase(key) like "CUSTOMOBJECT&custObjValue..DISPLAYNAME.TXT";
    quit;
  
    /* Add labels to the dataset */
    data work.&custObjAliasNm._fnl;
      set work.&custObjAliasNm._fnl;

      label
      %if %bquote(&custObjAliasNm) ne %str() and %bquote(&custObjLabel) ne  %str() %then
      %do;
        &custObjAliasNm._id="%trim(&custObjLabel) %sysfunc(sasmsg(sashelp.rskmgtsvr, Id, NOQUOTE))"
        &custObjAliasNm._rk="%trim(&custObjLabel) %sysfunc(sasmsg(sashelp.rskmgtsvr, Rk, NOQUOTE))"
        &custObjAliasNm._nm="%trim(&custObjLabel) %sysfunc(sasmsg(sashelp.rskmgtsvr, Nm, NOQUOTE))"
        &custObjAliasNm._desc="%trim(&custObjLabel) %sysfunc(sasmsg(sashelp.rskmgtsvr, Desc, NOQUOTE))"
        source_system_cd="%trim(&custObjLabel) %sysfunc(sasmsg(sashelp.rskmgtsvr, Source_System_Cd, NOQUOTE))"
      %end;
      %else %if %bquote(&custObjAliasNm) ne %str() and %bquote(&custObjLabel) eq %str() %then
      %do;
        &custObjAliasNm._id="%upcase(&custObjAliasNm) %sysfunc(sasmsg(sashelp.rskmgtsvr, Id, NOQUOTE))"
        &custObjAliasNm._rk="%upcase(&custObjAliasNm) %sysfunc(sasmsg(sashelp.rskmgtsvr, Rk, NOQUOTE))"
        &custObjAliasNm._nm="%upcase(&custObjAliasNm) %sysfunc(sasmsg(sashelp.rskmgtsvr, Nm, NOQUOTE))"
        &custObjAliasNm._desc="%upcase(&custObjAliasNm) %sysfunc(sasmsg(sashelp.rskmgtsvr, Desc, NOQUOTE))"
        source_system_cd="%upcase(&custObjAliasNm) %sysfunc(sasmsg(sashelp.rskmgtsvr, Source_System_Cd, NOQUOTE))"
      %end;
      
     %do y=1 %to 5 %by 1;
       %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CUSTOMUSER&y, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_CUSTOMUSER&y %then
       %do;
         customUser&y="%sysfunc(sasmsg(sashelp.rskmgtsvr, User_Name_&y, NOQUOTE))"
       %end;
       %else
       %do;
          customUser&y="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CUSTOMUSER&y, NOQUOTE))"
       %end;
     %end;

     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CREATOR, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_CREATOR %then
     %do;
       created_user_name="%sysfunc(sasmsg(sashelp.rskmgtsvr, Created_User_Name, NOQUOTE))"
     %end;
     %else
     %do;
       created_user_name="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CREATOR, NOQUOTE))"
     %end;
   
     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_EDITOR, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_EDITOR %then
     %do;
       editor_user_name="%sysfunc(sasmsg(sashelp.rskmgtsvr, Editor_User_Name, NOQUOTE))"
     %end;
     %else
     %do;
       editor_user_name="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_EDITOR, NOQUOTE))"
     %end;

     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_UPDATER, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_UPDATER %then
     %do;
       last_update_user_name="%sysfunc(sasmsg(sashelp.rskmgtsvr, Last_Update_User_Name, NOQUOTE))"
     %end;
     %else
     %do;
       last_update_user_name="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_UPDATER, NOQUOTE))"
     %end;
   
     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CHANGEREASON, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_CHANGEREASON %then
     %do;
       change_reason="%sysfunc(sasmsg(sashelp.rskmgtsvr, Change_Reason, NOQUOTE))"
     %end;
     %else
     %do;
       change_reason="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CHANGEREASON, NOQUOTE))"
     %end;

     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CREATEDDTTM, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_CREATEDDTTM %then
     %do;
       created_dttm="%sysfunc(sasmsg(sashelp.rskmgtsvr, Created_Dttm, NOQUOTE))"
     %end;
     %else
     %do;
       created_dttm="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CREATEDDTTM, NOQUOTE))"
     %end;
   
     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_VALIDFROMDDTTM, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_VALIDFROMDDTTM %then
     %do;
       valid_from_dttm="%sysfunc(sasmsg(sashelp.rskmgtsvr, Valid_From_Dttm, NOQUOTE))"
     %end;
     %else
     %do;
       valid_from_dttm="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_VALIDFROMDDTTM, NOQUOTE))"
     %end;
   
     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_VALIDTODTTM, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_VALIDTODTTM %then
     %do;
       valid_to_dttm="%sysfunc(sasmsg(sashelp.rskmgtsvr, Valid_To_Dttm, NOQUOTE))"
     %end;
     %else
     %do;
       valid_to_dttm="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_VALIDTODTTM, NOQUOTE))"
     %end;

     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_STATUSDTTM, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_STATUSDTTM %then
     %do;
       status_dttm="%sysfunc(sasmsg(sashelp.rskmgtsvr, Status_Dttm, NOQUOTE))"
     %end;
     %else
     %do;
       status_dttm="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_STATUSDTTM, NOQUOTE))"
     %end;
     
     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CREATEDFROMRK, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_CREATEDFROMRK %then
     %do;
       created_from_rk="%sysfunc(sasmsg(sashelp.rskmgtsvr, Created_From_Rk, NOQUOTE))"
     %end;
     %else
     %do;
       created_from_rk="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_CREATEDFROMRK, NOQUOTE))"
     %end;
   
     %if %sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_ACTIVEFLG, NOQUOTE))=CUSTOMOBJECT&custObjValue._OBJ_ACTIVEFLG %then
     %do;
       active_flg="%sysfunc(sasmsg(sashelp.rskmgtsvr, Active_Flg, NOQUOTE))"
     %end;
     %else
     %do;
       active_flg="%sysfunc(sasmsg(sashelp.rskmgtsvr, CUSTOMOBJECT&custObjValue._OBJ_ACTIVEFLG, NOQUOTE))"
     %end;

      %str(;)
    run;
  
    /* Create the &contentId.dat.&custObjAliasNm dataset in order to add to metadata */
    %if %eval(%sysfunc(exist(work.&custObjAliasNm._fnl)) = 1) %then
    %do;
      data &contentId.dat.&outdsn;
        if 0 then set work.&custObjAliasNm._fnl;
        stop;
      run;
    %end;
  
    /* Update the &contentId.dat table */
    proc append base=&contentId.dat.&outdsn data=work.&custObjAliasNm._fnl force;
    run;
 
    /* Load LASR Table and Authorizations into Metadata */
    %rskmgt_load_lasr_tbls(intblnm=&outdsn,
                           feature=CUST_OBJ_&custObjValue);  
      
    /* Remove work tables */
    proc datasets lib=work nolist;
      save
      
      %let y=1;
      %let contentIds=%trim(%qscan(%bquote(&rgf_content_ids),&y,%str(,)));
      %do %while (&contentIds ne );
  
        &contentIds._Custommessages &contentIds._sm_security_acl &contentIds._config_props
         
        %let y=%eval(&y+1);
        %let contentIds=%trim(%qscan(%bquote(&rgf_content_ids),&y,%str(,))); 
      %end;
      %str(;)
    quit;
    run;

    %continue:
  %end; /* End loop through all customObjects */
   
%put NOTE: {END: &sysmacroname};

%mend rskmgt_job_audit_custobj_data;