/*****************************************************************/
/* Copyright (c) 2019 by SAS Institute Inc., Cary, NC, USA.      */
/* All Rights Reserved.                                          */
/*****************************************************************/
/* NAME:     rskmgt_sp_unfold_dimensions.sas                     */
/* VERSION:  7.4                                                 */
/* DESCRIPTION: Expands an operational area to it's points and   */
/*              dimension paths                                  */
/* PRODUCT:  SAS Risk Management Server                          */
/*****************************************************************/

/* Unfold the dimensional points into each dimension and path */
%macro rskmgt_expand_dim_points(srclib=, /* the source data libname */
                                feature= /* the feature which we're filtering (e.g. OP_RISK_EVENT).
                                         This must match both a feature name in feature_dimensions, and
                                         be the BUSINESS_OBJECT_NM of the thing we're viewing. */
                               );
                        
%put NOTE: {START: &sysmacroname};

   %if (%upcase(&feature) = CAUSE or %upcase(&feature) = CONTROL) and (&dim_rk_column ne %str()) %then
   %do;
      %rskmgt_init_dimensionality(keep_dimension=&feature); 
   %end;
   %else
   %do;
     %rskmgt_init_dimensionality();   
   %end;
   
   /* Create a dim_point table w/ all possible dimensions per feature */
   proc sort data=&srclib..dim_point_l
       out=dim_point
         (keep=dim_point_rk %rskmgt_dim_keys_and_levels(feature=&feature));
       by dim_point_rk;
   run;   

%put NOTE: {END: &sysmacroname};

%mend rskmgt_expand_dim_points;

%macro rskmgt_sp_unfold_dimensions(srclib=, /* the source data libname */
                                in=, /* name of the input dataset */
                                rk_column=, /* name of the primary key column */
                                feature=, /* the feature which we're filtering (e.g. OP_RISK_EVENT).
                                            This must match both a feature name in feature_dimensions, and
                                            be the BUSINESS_OBJECT_NM of the thing we're viewing. */
                                business_object_nm=, /* business_object_nm must match the BUSINESS_OBJECT_NM in dim_point_mapping_l. */
                                dim_rk_column =, /* dimension key column used to join to the dim_point_l or dim_point_mapping_l tables */
                                                 /* used by the Event job to pull in the discovery location, cause and control type locations */
                                dim_rk_column_prefix = /* dimension key column prefix to use when renaming dimension columns in order */
                                                       /* to avoid naming conflicts when the &in table has multiple dimension point rk columns */                 
                                );

%put NOTE: {START: &sysmacroname};

 /* Add Incident Discovery and Cause Location Dimension Rks */
 %if (%upcase(&feature) = INCIDENT or %upcase(&feature) = CAUSE) and (&dim_rk_column ne %str()) %then
 %do;
   %rskmgt_expand_dim_points(srclib=&srclib,feature=&feature);

   /* Join the full set of dimension paths back to the dataset */
   proc sort tagsort data=dim_point (rename=(dim_point_rk=&dim_rk_column %rskmgt_dim_rename_keys_and_levels(feature=&feature, prefix=&dim_rk_column_prefix))); 
    by &dim_rk_column;
   
   proc sort tagsort data=&in; 
    by &dim_rk_column;

   data &in; 
     merge &in(in=left) dim_point; 
     by &dim_rk_column;
     if left;
   run;
 %end;
 %else
 %do;  
   /* OpRisk Area mappings */
   data iap_dpm(keep=dim_point_rk &rk_column);
      set &srclib..dim_point_mapping_l
         (rename=(business_object_rk=&rk_column));
      %if &business_object_nm eq %str() %then 
      %do;
        /* Handle scenarios where the cntl_features and */
        /* business_object_nm are not the same. */
        %if &feature=ASSESSMENT %then
        %do;
          if upcase(business_object_nm) = 'ANSWER_SHEET';
        %end;
        %else /* feature_dimensions and business_object_nm are the same */
        %do; 
          if upcase(business_object_nm) = "%upcase(&feature)";
        %end;
      %end;
      %else
      %do;
        if upcase(business_object_nm) = "%upcase(&business_object_nm)";
      %end;
   run;

   /* Control is Area Based so, we need to look in the Dim_Point_Mapping_L table */
   /* and then fetch the points from the Dim_Point_L table. */
   %if (%upcase(&feature) = CONTROL) and (&dim_rk_column ne %str()) %then
   %do;
      proc sort tagsort data=iap_dpm(rename=(dim_point_rk=&dim_rk_column)); 
        by &rk_column;
      run;

      proc sort tagsort data=&in;
        by &rk_column;
      run;
      
     proc sql undo_policy=none;
       create table &in as
       select a.*, b.&dim_rk_column
       from &in as a left join
            iap_dpm as b
       on a.&rk_column=b.&rk_column;
     quit;
      
     %rskmgt_expand_dim_points(srclib=&srclib,feature=&feature);

     /* Join the full set of dimension paths back to the dataset */
     proc sort tagsort data=dim_point(rename=(dim_point_rk=&dim_rk_column %rskmgt_dim_rename_keys_and_levels(feature=&feature, prefix=&dim_rk_column_prefix))); 
       by &dim_rk_column;
     run;

     proc sort tagsort data=&in; 
       by &dim_rk_column;
     run;
     
     proc sql undo_policy=none;
       create table &in(drop=dup_rk_col) as
       select a.*, b.*
       from &in as a left join
            dim_point(rename=(&dim_rk_column=dup_rk_col)) as b
       on a.&dim_rk_column=b.dup_rk_col;
     quit;
      
   %end;
   %else
   %do;
     proc sort tagsort data=iap_dpm; 
      by &rk_column;
     run;

     proc sort tagsort data=&in;
      by &rk_column;
     run;

     proc sql undo_policy=none;
       create table &in as
       select a.*, b.dim_point_rk
       from &in as a left join
            iap_dpm as b
       on a.&rk_column=b.&rk_column;
     quit;
     
     %rskmgt_expand_dim_points(srclib=&srclib,feature=&feature);
     
     /* Subset the points to only include dim_points in the &indsn */
     proc sql;
       create table work.dim_point as
       select distinct a.*
       from work.dim_point as a join
            &in as b
       on a.dim_point_rk=b.dim_point_rk;
     quit;
     
     /* If this customObject has no dim_points, stop processing dim_points. */
     %rskmgt_ds_exists(indsn=work.dim_point,
                      output_nobs=dim_point_cnt);

    %if &dim_point_cnt > 0 %then
    %do;      
      /* Subset the list of dim_point columns to include only the ones w/ values */
       /* check for missing numeric values */
       proc means data=work.dim_point noprint;
                  output out=checknum(drop=_:) n=;
       run;
   
       /* Transpose the list of variables into columns and keep only the dimensions w/ values */
       proc transpose data=checknum out=dims2 (drop=_label_ rename=(col1=numhere _name_=vname) where=(numhere ^=0));
       run;
   
       /* Put the list of dimensions w/ values into a list */
       proc sql noprint;
         select vname into : dims_w_vals separated by ' '
         from work.dims2;
       quit;
   
       /* Filter the dim_point table to include only the dimensions */
       /* for this feature and it must have a dim_point value for that node. */
       proc sort tagsort data=work.dim_point(keep=&dims_w_vals); 
         by dim_point_rk;
       run;

       proc sort tagsort data=&in; 
         by dim_point_rk;
       run;

       proc sql undo_policy=none;
         create table &in(drop=dup_rk_col) as
         select a.*, b.*
         from &in as a left join
              dim_point(rename=(dim_point_rk=dup_rk_col)) as b
         on a.dim_point_rk=b.dup_rk_col;
       quit;
     %end; 
   %end;
 %end;

%put NOTE: {END: &sysmacroname};
 
%mend rskmgt_sp_unfold_dimensions;
